package fsl.depict.shapefile;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.awt.image.ImageObserver;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.InputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import com.svcon.jdbf.DBFReader;
import com.svcon.jdbf.JDBField;
import fsl.depict.Depictable;
import fsl.depict.DepictableDisplayProperties;
import fsl.depict.ProductVisualInfo;
import fsl.util.DisplayCoordConverter;
import fsl.util.LatLong;
import fsl.util.Logger;
import fsl.util.ProductInfo;

/**
 * The shapefile depictable, used to encapsulate a single
 * frame of a hydrometeorological product in the form of
 * an ESRI Shapefile.
 *
 * @author Christopher Golden
 */
public class ShapefileDepictable extends Depictable {


    // Public Enumerated Types

    /**
     * Data types for properties.
     */
    public enum PropertyDataType { CHARACTER, INTEGER, FLOATING }


    // Protected Classes

    /**
     * Shape abstract base class, from which specific
     * shapes are derived.
     */
    protected abstract class GeoShape implements Serializable, Cloneable {


        // Protected Variables

        /**
         * Identifier.
         */
        protected int identifier;

        /**
         * List of property values for this shape; each
         * value goes with the property name at the
         * corresponding index in the enclosing class's
         * <code>propertyNames</code>.
         */
        protected Object[] properties;


        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param identifier Identifier.
         * @param properties List of property values for
         *                   this shape; each value goes
         *                   with the property name at
         *                   the corresponding index in
         *                   the enclosing class's
         *                   <code>propertyNames</code>.
         */
        public GeoShape(int identifier, Object[] properties) {
            this.identifier = identifier;
            this.properties = properties;
        }


        // Public Methods

        /**
         * Get the identifier.
         *
         * @return Identifier.
         */
        public int getIdentifier() {
            return identifier;
        }

        /**
         * Get the list of property values for this shape.
         * Each value goes with the property name at the
         * corresponding index in the enclosing class's
         * <code>propertyNames</code>.
         *
         * @return List of property values for this shape.
         */
        public Object[] getProperties() {
            return properties;
        }

        /**
         * Set the list of property values for this shape.
         * Each value goes with the property name at the
         * corresponding index in the enclosing class's
         * <code>propertyNames</code>.
         *
         * @param properties List of property values for
         *                   this shape.
         */
        public void setProperties(Object[] properties) {
            this.properties = properties;
        }
    }

    /**
     * Point shape.
     */
    protected class GeoPoint extends GeoShape {


        // Protected Variables

        /**
         * Coordinates.
         */
        protected LatLong coordinates;


        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param identifier  Identifier.
         * @param properties  List of property values for
         *                    this shape; each value goes
         *                    with the property name at
         *                    the corresponding index in
         *                    the enclosing class's
         *                    <code>propertyNames</code>.
         * @param coordinates Coordinates.
         */
        public GeoPoint(int identifier, Object[] properties,
                        LatLong coordinates) {
            super(identifier, properties);
            this.coordinates = coordinates;
        }


        // Public Methods

        /**
         * Get the coordinates.
         *
         * @return Coordinates.
         */
        public LatLong getCoordinates() {
            return coordinates;
        }
    }

    /**
     * Path shape, used for lines.
     */
    protected class GeoPath extends GeoShape {


        // Protected Variables

        /**
         * List of coordinates making up the line.
         */
        protected LatLong[] coordinates;


        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param identifier  Identifier.
         * @param properties  List of property values for
         *                    this shape; each value goes
         *                    with the property name at
         *                    the corresponding index in
         *                    the enclosing class's
         *                    <code>propertyNames</code>.
         * @param coordinates List of coordinates making
         *                    up the line.
         */
        public GeoPath(int identifier, Object[] properties,
                       LatLong[] coordinates) {
            super(identifier, properties);
            this.coordinates = coordinates;
        }


        // Public Methods

        /**
         * Get the list of coordinates making up the
         * line.
         *
         * @return List of coordinates.
         */
        public LatLong[] getCoordinates() {
            return coordinates;
        }
    }

    /**
     * Polygon shape, used for polygons.
     */
    protected class GeoPolygon extends GeoShape {


        // Protected Variables

        /**
         * List of polygons, each being made up of a
         * list of coordinates. A line segment will
         * be assumed between the last and first
         * coordinates in each part, forming a ring.
         * Thus, each part's last coordinate should
         * not be a repeat of the first. Clockwise
         * rings each form the outside edge of a
         * polygon that makes up part of the total
         * polygon; counterclockwise rings form holes
         * in the polygon.
         */
        protected LatLong[][] coordinates;


        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param identifier  Identifier.
         * @param properties  List of property values for
         *                    this shape; each value goes
         *                    with the property name at
         *                    the corresponding index in
         *                    the enclosing class's
         *                    <code>propertyNames</code>.
         * @param coordinates List of polygons, each being
         *                    made up of a list of
         *                    coordinates. A line segment
         *                    will be assumed between the
         *                    last and first coordinates
         *                    in each part, forming a ring.
         *                    Thus, each part's last
         *                    coordinate should not be a
         *                    repeat of the first.
         *                    Clockwise rings each form the
         *                    outside edge of a polygon
         *                    that makes up part of the
         *                    total polygon; counterclockwise
         *                    rings form holes in the polygon.
         */
        public GeoPolygon(int identifier, Object[] properties,
                          LatLong[][] coordinates) {
            super(identifier, properties);
            this.coordinates = coordinates;
        }


        // Public Methods

        /**
         * Get the list of polygons, each being made up
         * of a list of coordinates. A line segment will
         * be assumed between the last and first
         * coordinates in each part, forming a ring.
         * Thus, each part's last coordinate should not
         * be a repeat of the first. Clockwise rings each
         * form the outside edge of a polygon that makes
         * up part of the total polygon; counterclockwise
         * rings form holes in the polygon.
         *
         * @return List of rings making up the polygon.
         */
        public LatLong[][] getCoordinates() {
            return coordinates;
        }
    }

    /**
     * Data record containing a shape for drawing and the
     * list of property values associated with the shape.
     */
    protected class DrawingShape {


        // Protected Variables

        /**
         * Drawable shape.
         */
        protected Shape shape;

        /**
         * List of property values for this shape; each
         * value goes with the property name at the
         * corresponding index in the enclosing class's
         * <code>propertyNames</code>.
         */
        protected Object[] properties;


        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param shape      Drawable shape.
         * @param properties List of property values for
         *                   this shape; each value goes
         *                   with the property name at
         *                   the corresponding index in
         *                   the enclosing class's
         *                   <code>propertyNames</code>.
         */
        public DrawingShape(Shape shape, Object[] properties) {
            this.shape = shape;
            this.properties = properties;
        }


        // Public Methods

        /**
         * Get the drawable shape.
         *
         * @return Drawable shape.
         */
        public Shape getShape() {
            return shape;
        }

        /**
         * Get the list of property values for this shape.
         * Each value goes with the property name at the
         * corresponding index in the enclosing class's
         * <code>propertyNames</code>.
         *
         * @return List of property values for this shape.
         */
        public Object[] getProperties() {
            return properties;
        }
    }

    /**
     * Data record containing a point that may be a candidate
     * for having associated text drawn alongside it.
     */
    protected class AnnotatedPoint {


        // Protected Variables

        /**
         * Location of the point.
         */
        protected Point point;

        /**
         * Associated text.
         */
        protected String text;


        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param point Location of the point.
         * @param text   Associated text.
         */
        public AnnotatedPoint(Point point, String text) {
            this.point = point;
            this.text = text;
        }


        // Public Methods

        /**
         * Get the location of the point.
         *
         * @return Location of the point.
         */
        public Point getPoint() {
            return point;
        }

        /**
         * Get the associated text.
         *
         * @return Associated text.
         */
        public String getText() {
            return text;
        }
    }

    /**
     * Data record containing a two-dimensional shape that
     * may be a candidate for being annotated with text.
     */
    protected class AnnotatedShape implements Comparable<AnnotatedShape> {


        // Protected Variables

        /**
         * Associated text.
         */
        protected String text;

        /**
         * Relative priority for drawing annotations.
         */
        protected int priority;

        /**
         * List of center cells of previously drawn text
         * labels for this shape.
         */
        protected ArrayList<Point> centerCells = new ArrayList<Point>();


        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param text     Associated text.
         * @param priority Relative priority for drawing
         *                 annotations.
         */
        public AnnotatedShape(String text, int priority) {
            this.text = text;
            this.priority = priority;
        }


        // Public Methods

        /**
         * Get the associated text.
         *
         * @return Associated text.
         */
        public String getText() {
            return text;
        }

        /**
         * Get the relative priority for annotation
         * drawing.
         *
         * @return Relative priority for annotation
         *         drawing.
         */
        public int getPriority() {
            return priority;
        }

        /**
         * Set the relative priority for annotation
         * drawing.
         *
         * @param priority New relative priority.
         */
        public void setPriority(int priority) {
            this.priority = priority;
        }

        /**
         * Get the cells in which this annotated shape has
         * already had its text drawn.
         *
         * @return Cells in which this annotated shape has
         *         already had its text drawn.
         */
        public ArrayList<Point> getAnnotatedCells() {
            return centerCells;
        }

        /**
         * Add the specified cell to the list of cells in
         * which this annotated shape has already had its
         * text drawn.
         *
         * @param Cell to be added.
         */
        public void addToAnnotatedCells(Point cell) {
            centerCells.add(cell);
        }

        /**
         * Compare this object with the specified object for
         * order.
         *
         * @param  other Object to which this one is to be
         *               compared.
         * @return Negative integer, zero, or positive integer
         *         if this object is respectively less than,
         *         equal to, or greater than the specified
         *         object, respectively.
         */
        public int compareTo(AnnotatedShape other) {
            return other.priority - priority;
        }
    }

    /**
     * Data record containing a path that may be a candidate
     * for being annotated with text.
     */
    protected class AnnotatedPath extends AnnotatedShape {


        // Protected Variables

        /**
         * Points defining the path.
         */
        protected ArrayList<Point> path;

        /**
         * Set of lists of paths for cells in which this
         * path appears.
         */
        protected HashSet<ArrayList<AnnotatedPath>> cells;


        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param path     Points defining the path.
         * @param text     Associated text.
         * @param priority Relative priority for annotation
         *                 drawing.
         * @param cells    Cells in which this annotated path
         *                 is present.
         */
        public AnnotatedPath(ArrayList<Point> path, String text, int priority,
                             HashSet<ArrayList<AnnotatedPath>> cells) {
            super(text, priority);
            this.path = path;
            this.cells = cells;
        }


        // Public Methods

        /**
         * Get the path.
         *
         * @return Path.
         */
        public ArrayList<Point> getPath() {
            return path;
        }

        /**
         * Get the set of lists of paths for cells in which
         * this path appears.
         *
         * @return Set of lists of paths for cells in which
         *         this path appears.
         */
        public HashSet<ArrayList<AnnotatedPath>> getCells() {
            return cells;
        }
    }

    /**
     * Data record containing a polygon that may be a
     * candidate for being annotated with text.
     */
    protected class AnnotatedPolygon extends AnnotatedShape {


        // Protected Variables

        /**
         * Bounding rectangle clipped to the display.
         */
        protected Rectangle boundsRect;

        /**
         * General path making up the polygon's boundary.
         */
        protected GeneralPath path;

        /**
         * Center point of the polygon.
         */
        protected Point centerPoint;



        // Public Constructors

        /**
         * Construct a standard instance.
         *
         * @param text        Associated text.
         * @param path        General path making up the
         *                    polygon's boundary.
         * @param priority    Relative priority for drawing
         *                    annotations.
         * @param boundsRect  Bounding rectangle clipped to
         *                    the display.
         * @param centerPoint Point that is the center of
         *                    the polygon when clipped to
         *                    the display.
         */
        public AnnotatedPolygon(String text, GeneralPath path, int priority,
                                Rectangle boundsRect, Point centerPoint) {
            super(text, priority);
            this.boundsRect = boundsRect;
            this.path = path;
            this.centerPoint = centerPoint;
        }


        // Public Methods

        /**
         * Get the bounding rectangle of the polygon.
         *
         * @return Bounding rectangle of the polygon.
         */
        public Rectangle getBounds() {
            return boundsRect;
        }

        /**
         * Get the path making up the boundaries
         * of the polygon.
         *
         * @return Path making up the boundaries
         *         of the polygon.
         */
        public GeneralPath getPath() {
            return path;
        }

        /**
         * Get the center point of the polygon.
         *
         * @return Center point of the polygon.
         */
        public Point getCenterPoint() {
            return centerPoint;
        }
    }


    // Protected Static Constants

    /**
     * Shapefile point record type.
     */
    protected static final int POINT_TYPE = 1;

    /**
     * Shapefile polyline record type.
     */
    protected static final int POLYLINE_TYPE = 3;

    /**
     * Shapefile polygon record type.
     */
    protected static final int POLYGON_TYPE = 5;

    /**
     * Default starting color for paths.
     */
    protected static final Color COLOR = new Color(155, 155, 155);

    /**
     * Standard basic stroke that yields a solid thin line
     * when used.
     */
    protected static final BasicStroke[] SOLID_STROKE = {
        new BasicStroke(System.getProperty("fxc.display.smoothing", "false").
                        toLowerCase().equals("true") ? 1.0f : 0.0f),
        new BasicStroke(2.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL),
        new BasicStroke(3.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL),
        new BasicStroke(4.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL)
    };

    /**
     * Minimum length in pixels of the visible bounding
     * rectangle of a path required to make that path a
     * candidate for displaying associated text.
     */
    protected static final double MINIMUM_PATH_LENGTH_CELL_MULTIPLIER = 4;

    /**
     * Minimum number of cells that should be between
     * labels for the same path when drawn.
     */
    protected static final int ANNOTATION_SPACING_FOR_SAME_PATH = 8;

    /**
     * Minimum number of cells that should be between
     * labels for the same polygon when drawn.
     */
    protected static final int ANNOTATION_SPACING_FOR_SAME_POLYGON = 24;

    /**
     * Vertical offset in pixels between an annotated
     * point and the bottom of its annotating text.
     */
    protected static final int POINT_ANNOTATION_VERTICAL_OFFSET = 2;

    /**
     * Above direction.
     */
    protected static final int ABOVE = 0;

    /**
     * Below direction.
     */
    protected static final int RIGHT = 1;

    /**
     * Left direction.
     */
    protected static final int BELOW = 2;

    /**
     * Right direction.
     */
    protected static final int LEFT = 3;


    // Protected Variables

    /**
     * List of shapes making up the depictable.
     */
    protected ArrayList<GeoShape> shapes;

    /**
     * Array of property names.
     */
    protected String[] propertyNames;

    /**
     * Hash map pairing property names with the indices
     * at which those names are found in <code>
     * propertyNames</code>.
     */
    protected HashMap<String, Integer> propertyIndicesByNames;

    /**
     * Timestamp indicating creation time.
     */
    protected long creationTime;

    /**
     * List of default property mappings.
     */
    protected ArrayList<ShapefileDisplayProperties.Property> propertyMappings;

    /**
     * Cached list of drawable shapes.
     */
    protected transient ArrayList<DrawingShape> cachedDrawables = null;

    /**
     * Cached grid of lists indicating which points are
     * in which cells.
     */
    protected transient ArrayList<AnnotatedPoint>[][] cachedPointsInCells = null;

    /**
     * Cached grid of lists indicating which paths are
     * in which cells.
     */
    protected transient ArrayList<AnnotatedPath>[][] cachedPathsInCells = null;

    /**
     * Cached grid of lists indicating which polygons'
     * centroids are in which cells.
     */
    protected transient ArrayList<AnnotatedPolygon>[][] cachedCentroidsInCells = null;

    /**
     * Cached grid indicating which polygons are in
     * which cells.
     */
    protected transient AnnotatedPolygon[][] cachedPolygonsInCells = null;

    /**
     * Cached list of polygons.
     */
    protected transient ArrayList<AnnotatedPolygon> cachedPolygons = null;

    /**
     * Cached label high-priority text strings; this
     * cache is display context-sensitive, and is thus
     * transient.
     */
    protected transient ArrayList<String> cachedHiPriLabelStrings = null;

    /**
     * Cached label high-priority angles; this cache is
     * display context-sensitive, and is thus transient.
     * Each angle is for the label at the corresponding
     * index in <code>cachedHiPriLabelStrings</code>.
     */
    protected transient ArrayList<Double> cachedHiPriLabelAngles = null;

    /**
     * Cached label high-priority center points; this
     * cache is display context-sensitive, and is thus
     * transient. Each point is for the label at the
     * corresponding index in <code>
     * cachedHiPriLabelStrings</code>.
     */
    protected transient ArrayList<Point> cachedHiPriLabelPoints = null;

    /**
     * Cached label high-priority set of cells occupied
     * by the label at the corresponding index in <code>
     * cachedHiPriLabelStrings</code>. This cache is
     * display context-sensitive, and is thus transient.
     */
    protected transient ArrayList<HashSet<Point>> cachedHiPriLabelCells = null;

    /**
     * Cached label high-priority points that are to be
     * visually emphasized when drawing the label found
     * at the corresponding index of <code>
     * cachedHiPriLabelStrings</code>. If the value at
     * an index is <code>null</code>, then no point is
     * to be visually emphasized for that label. This
     * cache is display context-sensitive, and is thus
     * transient.
     */
    protected transient ArrayList<Point> cachedHiPriLabelVisualEmphasisPoints = null;

    /**
     * Cached label lower-priority text strings; this
     * cache is display context-sensitive, and is thus
     * transient.
     */
    protected transient ArrayList<String> cachedOtherLabelStrings = null;

    /**
     * Cached label lower-priority angles; this cache is
     * display context-sensitive, and is thus transient.
     * Each angle is for the label at the corresponding
     * index in <code>cachedOtherLabelStrings</code>.
     */
    protected transient ArrayList<Double> cachedOtherLabelAngles = null;

    /**
     * Cached label lower-priority center points; this
     * cache is display context-sensitive, and is thus
     * transient. Each point is for the label at the
     * corresponding index in <code>
     * cachedOtherLabelStrings</code>.
     */
    protected transient ArrayList<Point> cachedOtherLabelPoints = null;

    /**
     * Cached label lower-priority set of cells occupied
     * by the label at the corresponding index in <code>
     * cachedOtherLabelStrings</code>. This cache is
     * display context-sensitive, and is thus transient.
     */
    protected transient ArrayList<HashSet<Point>> cachedOtherLabelCells = null;

    /**
     * Cached label font.
     */
    protected transient Font cachedFont = null;


    // Public Constructors

    /**
     * Create an empty instance.
     */
    public ShapefileDepictable() {
    } 

    /**
     * Create an instance from the specified file for the
     * specified product.
     *
     * @param key         Key of the depictable.
     * @param file        File from which the depictable is
     *                    to be created.
     * @param compression Compression applied to the data;
     *                    this holds one of the values
     *                    defined as compression types in
     *                    <code>Packer</code>.
     * @param productInfo Info about the product for which
     *                    this depictable is being created.
     * @see fsl.util.Packer
     */
    @SuppressWarnings(value={"unchecked"})
    public ShapefileDepictable(long key, File file, int compression,
                               ProductInfo productInfo) throws InterruptedException {
        super(key, file, compression, productInfo);

        // Load the shapefile and convert it to depictable
        // shapes.
        try {

            // Determine the creation time as the current
            // time if reading from a URL, or the file
            // modification time if reading from a file.
            if (file.getPath().startsWith("http://"))
                creationTime = System.currentTimeMillis();
            else
                creationTime = file.lastModified();

            // Get the bytes from the ZIP file.
            File zipFile = file;
            byte[] zipBytes = readFileOrURL(zipFile);
            if (zipBytes == null)
                throw new InterruptedException("Could not read the file " + zipFile + ".");

            // Iterate through the ZIP archive, skipping
            // all files found therein except for the
            // SHP and DBF files, which are read in for
            // the shape and property data, respectively.
            ZipInputStream zipStream = new ZipInputStream(new ByteArrayInputStream(zipBytes));
            ZipEntry zipEntry = null;
            boolean shpDone = false, dbfDone = false;
            ArrayList<Object[]> propertyValues = null;
            shapes = new ArrayList<GeoShape>();
            while (((zipEntry = zipStream.getNextEntry()) != null) &&
                   ((shpDone == false) || dbfDone == false)) {
                String name = zipEntry.getName().toLowerCase();
                if (name.endsWith(".shp")) {
                    long time = System.currentTimeMillis();
                    ByteArrayInputStream inStream = getByteStream(zipStream, (int)
                                                                  zipEntry.getSize());
                    readSHP(inStream, zipEntry.getSize());
                    inStream.close();
                    Logger.logVerbose("Reading of key " + key + "'s SHP file done, took " +
                                      ((System.currentTimeMillis() - time) / 1000L) + "." +
                                      ((System.currentTimeMillis() - time) % 1000L) +
                                      " seconds.");
                    shpDone = true;
                } else if (name.endsWith(".dbf")) {
                    long time = System.currentTimeMillis();
                    ByteArrayInputStream inStream = getByteStream(zipStream, (int)
                                                                  zipEntry.getSize());
                    propertyValues = readDBF(inStream, zipEntry.getSize());
                    inStream.close();
                    Logger.logVerbose("Reading of key " + key + "'s DBF file done, took " +
                                      ((System.currentTimeMillis() - time) / 1000L) + "." +
                                      ((System.currentTimeMillis() - time) % 1000L) +
                                      " seconds.");
                    dbfDone = true;
                }
                zipStream.closeEntry();
            }
            zipStream.close();
            zipBytes = null;

            // Add the property values to the shapes
            // as appropriate.
            for (int j = 0, k = -1; j < shapes.size(); j++) {
                if ((j == 0) || (shapes.get(j - 1).getIdentifier() !=
                                 shapes.get(j).getIdentifier()))
                    k++;
                shapes.get(j).setProperties(propertyValues.get(k));
            }
        } catch (Exception e) {
            Logger.logBug("A problem occurred while converting the shapefile.", e);
        }
        System.gc();
    }


    // Public Methods

    /**
     * Determine whether this depictable is likely to
     * paint annotations or not.
     *
     * @return True if the depictable may wish to paint
     *         annotations, false otherwise. If false,
     *         the depictable is guaranteeing that it
     *         is not going to want to paint annotations
     *         that it is concerned will overwrite or be
     *         overwritten by other annotations.
     */
    public boolean hasAnnotations() {
        return true;
    }

    /**
     * Receive notification that the annotation sizes
     * may have changed.
     */
    public void annotationSizesChanged() {
        cachedHiPriLabelStrings = null;
        cachedHiPriLabelAngles = null;
        cachedHiPriLabelPoints = null;
        cachedHiPriLabelCells = null;
    }

    /**
     * Draw the depictable using the cached display
     * parameters and the specified graphics object.
     *
     * @param  graphics       Graphics object to be used.
     * @param  visualInfo     Visual information about
     *                        this map or product, if any.
     * @param  depictProps    Display properties to be
     *                        used, if any.
     * @param  observer       Image observer that is to be
     *                        notified if a paint cannot
     *                        be completed because an
     *                        image has not yet finished
     *                        loading.
     * @param  annotationGrid Annotation grid indicating
     *                        which cells are occupied
     *                        and which are not. This
     *                        grid is supplied here in
     *                        case any preparation is to
     *                        be done within this method
     *                        for the subsequent calls to
     *                        <code>paintPrimaryAnnotation()
     *                        </code> and <code>
     *                        paintOtherAnnotation()</code>.
     * @return If painting could not be completed because
     *         a required image has not yet finished
     *         loading, the image that is still loading
     *         is returned. Otherwise, <code>null</code>
     *         is returned. For shapefile depictables,
     *         <code>null</code> is always returned, as no
     *         images are used.
     */
    @SuppressWarnings("unchecked")
    public Image paint(Graphics graphics,
                       ProductVisualInfo visualInfo,
                       DepictableDisplayProperties depictProps,
                       ImageObserver observer,
                       AnnotationGrid annotationGrid) {

        // Ensure that the display is showing a two-
        // dimensional space.
        if ((width < 1) || (height < 1))
            return null;

        // If there is nothing to paint, do nothing.
        if (shapes == null)
            return null;


        /*
        long startTime = System.currentTimeMillis();
        */




        // If the display properties are empty, fill them
        // in.  Otherwise, if the display properties pro-
        // vided are older than the ones just read in,
        // incorporate the old ones into the new one and
        // replace the old one with the new one.
        ShapefileDisplayProperties shapefileProperties =
            (ShapefileDisplayProperties) depictProps;
        if (shapefileProperties.isPopulated() == false)
            shapefileProperties.populatePropertyMappings(creationTime,
                                                         propertyMappings, -1);
        else if (shapefileProperties.getCreationTime() < creationTime) {
            ShapefileDisplayProperties newProperties =
                new ShapefileDisplayProperties(creationTime, propertyMappings);
            if (shapefileProperties.isBasicallySameAs(newProperties) == false) {
                newProperties.incorporateRelevantMappings(shapefileProperties);
                shapefileProperties.
                    populatePropertyMappings(creationTime,
                                             newProperties.getPropertyMappings(),
                                             newProperties.getCurrentPropertyMappingsIndex());
            }                
        }

        // Deterine which property mapping will be used to
        // govern the drawing of the shapes.
        int propertyIndex = shapefileProperties.getCurrentPropertyMappingsIndex();
        ShapefileDisplayProperties.Property property = null;
        int propertyNameIndex = -1;
        if (propertyIndex != -1) {
            property = shapefileProperties.getPropertyMapping(propertyIndex);
            propertyNameIndex = propertyIndicesByNames.get(property.getName());
        }

        // Create the coordinate converter.
        DisplayCoordConverter converter = new DisplayCoordConverter();
        converter.setWorldCoordTransformer(coordXfm);
        converter.setDisplayWidth(width);
        converter.setDisplayHeight(height);
        float[] scaleInfo = coordXfm.getParameters();
        converter.setScaleWidth(scaleInfo[10]);
        converter.setScaleHeight(scaleInfo[11]);
        converter.setUpperLeftX(scaleInfo[8]);
        converter.setUpperLeftY(scaleInfo[9]);
        converter.setZoom(zoom);
        converter.setZoomX(zoomX);
        converter.setZoomY(zoomY);

        // Determine how many cells exist in the annotation
        // grid, and the horizontal and vertical dimensions
        // of each cell (except the last row and column,
        // which might have smaller dimensions).
        int numberOfCellsHoriz = annotationGrid.getNumberOfCellsHorizontally();
        int cellWidth = annotationGrid.getCellWidth();
        int numberOfCellsVert = annotationGrid.getNumberOfCellsVertically();
        int cellHeight = annotationGrid.getCellHeight();

        // Determine whether or not drawable shapes have
        // been cached from a previous paint, and if not,
        // get ready to create the caches.
        boolean drawablesDone = (cachedDrawables != null);
        if (drawablesDone == false) {
            cachedDrawables = new ArrayList<DrawingShape>();
            cachedPointsInCells = new ArrayList[numberOfCellsHoriz][numberOfCellsVert];
            cachedPathsInCells = new ArrayList[numberOfCellsHoriz][numberOfCellsVert];
            cachedCentroidsInCells = new ArrayList[numberOfCellsHoriz][numberOfCellsVert];
            for (int y = 0; y < numberOfCellsVert; y++) {
                for (int x = 0; x < numberOfCellsHoriz; x++) {
                    cachedPointsInCells[x][y] = new ArrayList<AnnotatedPoint>();
                    cachedPathsInCells[x][y] = new ArrayList<AnnotatedPath>();
                    cachedCentroidsInCells[x][y] = new ArrayList<AnnotatedPolygon>();
                }
            }
            cachedPolygons = new ArrayList<AnnotatedPolygon>();
            cachedPolygonsInCells = new AnnotatedPolygon[numberOfCellsHoriz][numberOfCellsVert];
        }

        // Determine whether or not text placement has
        // been determined during a previous paint.  If no
        // text is to be painted, consider it done.  If it
        // is not done, get ready to create the caches to
        // hold the placement information.
        boolean textPlacementDone = ((cachedHiPriLabelStrings != null) ||
                                     (shapefileProperties.isShowingLabels() == false));
        if (textPlacementDone == false) {
            cachedHiPriLabelStrings = new ArrayList<String>();
            cachedHiPriLabelAngles = new ArrayList<Double>();
            cachedHiPriLabelPoints = new ArrayList<Point>();
            cachedHiPriLabelCells = new ArrayList<HashSet<Point>>();
            cachedHiPriLabelVisualEmphasisPoints = new ArrayList<Point>();
            cachedOtherLabelStrings = new ArrayList<String>();
            cachedOtherLabelAngles = new ArrayList<Double>();
            cachedOtherLabelPoints = new ArrayList<Point>();
            cachedOtherLabelCells = new ArrayList<HashSet<Point>>();
        }
        ArrayList<AnnotatedPoint> annotatedPoints = null;
        HashMap<String, LinkedList<ArrayList<Point>>> pathsForNames = null;
        HashMap<String, ArrayList<AnnotatedPolygon>> polygonsForName = null;

        // If drawables have not been cached, then create
        // them now.
        if (drawablesDone == false) {

            // Create a list to hold the points that may be
            // annotated.
            annotatedPoints = new ArrayList<AnnotatedPoint>();

            // Create a hash table to hold lists of paths
            // that have the same name, and another hash
            // table doing the same thing for polygons.
            pathsForNames = new HashMap<String, LinkedList<ArrayList<Point>>>();
            polygonsForName = new HashMap<String, ArrayList<AnnotatedPolygon>>();

            // Get the list of annotation constant strings
            // and the list of annotation variable names,
            // to be used to construct each annotation.
            String[] annotationConstants = ((ShapefileVisualInfo) visualInfo).getConstants();
            String[] annotationVariables = ((ShapefileVisualInfo) visualInfo).getVariables();

            // Iterate through the shapes, creating each one
            // in turn.  While doing so, determine which ones
            // are candidates for having any associated text
            // drawn over them as well.
            Rectangle displayRect = new Rectangle(width, height);
            for (int j = 0; j < shapes.size(); j++) {

                // Process the shape for drawing.
                GeoShape shape = shapes.get(j);
                if (shape instanceof GeoPoint) {

                    // Construct a circle with its center at
                    // the point's location, and add it to the
                    // list of shapes to be drawn it if it is
                    // at least partly showing.
                    GeoPoint geoPoint = (GeoPoint) shape;
                    LatLong coordinates = geoPoint.getCoordinates();
                    Point point = converter.convert(coordinates);
                    Ellipse2D.Float ellipse = new Ellipse2D.Float(point.x - 2,
                                                                  point.y - 2, 4, 4);
                    if (ellipse.intersects(displayRect) == false)
                        continue;
                    cachedDrawables.add(new DrawingShape(ellipse, shape.getProperties()));

                    // Determine the text that goes with this
                    // point; if it has no text, do nothing
                    // more with it.
                    String text = getAnnotationTextForShape(annotationConstants,
                                                            annotationVariables,
                                                            geoPoint);
                    if (text.length() == 0)
                        continue;

                    // Add this point to the list of points
                    // that may be annotated.
                    annotatedPoints.add(new AnnotatedPoint(point, text));
                } else if (shape instanceof GeoPath) {

                    // Construct a path with a point for each
                    // of the original path's points.
                    GeoPath geoPath = (GeoPath) shape;
                    LatLong[] coordinates = geoPath.getCoordinates();
                    ArrayList<Point> points = new ArrayList<Point>();
                    GeneralPath path = new GeneralPath(GeneralPath.WIND_NON_ZERO,
                                                       coordinates.length);
                    Point lastPoint = null;
                    for (int k = 0; k < coordinates.length; k++) {
                        Point thisPoint = converter.convert(coordinates[k]);
                        if (thisPoint.equals(lastPoint))
                            continue;
                        lastPoint = thisPoint;
                        points.add(thisPoint);
                        if (k == 0)
                            path.moveTo(thisPoint.x, thisPoint.y);
                        else
                            path.lineTo(thisPoint.x, thisPoint.y);
                    }

                    // Do nothing more with the constructed
                    // path if is is not showing even partly
                    // on the display. Otherwise, add it to
                    // the list of shapes to be drawn.
                    if (path.intersects(displayRect) == false)
                        continue;
                    cachedDrawables.add(new DrawingShape(path, shape.getProperties()));

                    // Determine the text that goes with this
                    // path; if it has no text, do nothing
                    // more with it.
                    String text = getAnnotationTextForShape(annotationConstants,
                                                            annotationVariables,
                                                            geoPath);
                    if (text.length() == 0)
                        continue;

                    // If the path is the first one to have
                    // this text, create a list of paths with
                    // this name and put this one in it.
                    // Otherwise, see if it can be joined at
                    // one or both ends with previously re-
                    // corded paths with the same text, and
                    // if so, remove any it is joined with.
                    LinkedList<ArrayList<Point>> paths = pathsForNames.get(text);
                    if (paths == null) {
                        paths = new LinkedList<ArrayList<Point>>();
                        pathsForNames.put(text, paths);
                    } else {

                        // Try joining with the start of the
                        // path first, then the end of the
                        // path.
                        for (int end = 0; end < 2; end++) {

                            // Iterate through the paths with
                            // the same name, attempting to
                            // join the chosen end  of this
                            // path with either end of the
                            // other paths.
                            Point thisPoint = points.get(end != 0 ? points.size() - 1 : 0);
                            int joinWith = paths.size();
                            boolean joinAtStart = false;
                            for (int k = 0; k < paths.size(); k++) {

                                // See if this path joins
                                // with either end of the
                                // other path.
                                Point point1 = paths.get(k).get(0);
                                Point point2 = paths.get(k).get(paths.get(k).size() - 1);
                                for (int m = 0; m < 2; m++) {

                                    // If the difference
                                    // between the X and Y
                                    // coordinates of the
                                    // chosen ends of the
                                    // two paths is no more
                                    // than 1, these two
                                    // ends should be joined.
                                    int diffX = thisPoint.x - (m % 2 == 1 ? point2.x : point1.x);
                                    int diffY = thisPoint.y - (m % 2 == 1 ? point2.y : point1.y);
                                    if ((diffX >= -1) && (diffX <= 1) &&
                                        (diffY >= -1) && (diffY <= 1)) {
                                        joinWith = k;
                                        joinAtStart = (m % 2 == 0);
                                        break;
                                    }
                                }
                                if (joinWith < paths.size())
                                    break;
                            }

                            // If joining should occur, do
                            // it now.  First reverse the
                            // other path's points if the
                            // join requires it; then pre-
                            // pend or append it to this
                            // path.  Also remove the other
                            // path, as it is no longer a
                            // separate entity.
                            if (joinWith < paths.size()) {
                                ArrayList<Point> otherList = paths.remove(joinWith);
                                if ((joinAtStart && (end == 0)) ||
                                    (joinAtStart == false) && (end != 0)) {
                                    Collections.reverse(otherList);
                                }
                                points.addAll(end != 0 ? points.size() : 0, otherList);
                            }
                        }
                    }

                    // Add the new path to the list of paths
                    // with this name.
                    paths.add(points);
                } else {

                    // Construct each ring as a closed path.
                    // Counterclockwise paths will be
                    // treated as holes in the polygon that
                    // results, as they should be.
                    GeoPolygon geoPoly = (GeoPolygon) shape;
                    LatLong[][] coordinates = geoPoly.getCoordinates();
                    int numCoordinates = 0;
                    for (int k = 0; k < coordinates.length; k++)
                        numCoordinates += coordinates[k].length;
                    GeneralPath path = new GeneralPath(GeneralPath.WIND_EVEN_ODD,
                                                       numCoordinates);
                    ArrayList<Point> points = new ArrayList<Point>();
                    for (int k = 0; k < coordinates.length; k++) {
                        Point firstPoint = null, lastPoint = null;
                        for (int m = 0; m < coordinates[k].length; m++) {
                            Point thisPoint = converter.convert(coordinates[k][m]);
                            if (thisPoint.equals(lastPoint))
                                continue;
                            points.add(thisPoint);
                            lastPoint = thisPoint;
                            if (m == 0) {
                                firstPoint = thisPoint;
                                path.moveTo(thisPoint.x, thisPoint.y);
                            } else
                                path.lineTo(thisPoint.x, thisPoint.y);
                        }
                        path.closePath();
                        points.add(firstPoint);
                    }

                    // Do nothing more with the constructed
                    // path if is is not showing even partly
                    // on the display; otherwise, add it to
                    // the list of shapes to be drawn.
                    if (path.intersects(displayRect) == false)
                        continue;
                    cachedDrawables.add(new DrawingShape(path, shape.getProperties()));

                    // Determine the text that goes with this
                    // polygon; if it has no text, do nothing
                    // more with it.
                    String text = getAnnotationTextForShape(annotationConstants,
                                                            annotationVariables,
                                                            geoPoly);
                    if (text.length() == 0)
                        continue;

                    // Determine the visible part of the po-
                    // lygon's bounding rectangle, and calcu-
                    // late its area.
                    Rectangle polygonRect = path.getBounds().intersection(displayRect);
                    int area = polygonRect.width * polygonRect.height;

                    // Get the centroid of the polygon, and
                    // create a record for it.  Instead of
                    // the centroid, use the center of the
                    // bounding rectangle if the polygon is
                    // quite small.
                    Point centerPoint = ((polygonRect.width < 4) || (polygonRect.height < 4) ?
                                         new Point(polygonRect.x + (polygonRect.width / 2),
                                                   polygonRect.y + (polygonRect.height / 2)) :
                                         getCentroidOfPolygon(points));
                    AnnotatedPolygon polygon = new AnnotatedPolygon(text, path, area,
                                                                    polygonRect,
                                                                    centerPoint);

                    // Add the polygon to the list of poly-
                    // gons with the same text.
                    ArrayList<AnnotatedPolygon> polygonList = polygonsForName.get(text);
                    if (polygonList == null) {
                        polygonList = new ArrayList<AnnotatedPolygon>();
                        polygonsForName.put(text, polygonList);
                    }
                    polygonList.add(polygon);
                }
            }
        }

        // Set the pen width as specified by the display
        // parameters, and make lines and text smoother
        // if appropriate.
        ((Graphics2D) graphics).setStroke(SOLID_STROKE[0]);
        Object antiAliasingHint = null;
        if (System.getProperty("fxc.display.smoothing", "false").toLowerCase().equals("true")) {
            antiAliasingHint = ((Graphics2D) graphics).
                getRenderingHint(RenderingHints.KEY_ANTIALIASING);
            ((Graphics2D) graphics).setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                                                     RenderingHints.VALUE_ANTIALIAS_ON);
        }

        // Iterate through the drawable shapes, drawing
        // each in turn.
        for (int j = 0; j < cachedDrawables.size(); j++) {
            DrawingShape drawable = cachedDrawables.get(j);

            // Determine whether the shape should be drawn
            // or not, and if so, what its color should be,
            // then set the color.  If no property mapping
            // exists, go with the standard color.
            Color shapeColor = color;
            if (property != null) {
                Object value = drawable.getProperties()[propertyNameIndex];
                if (property instanceof ShapefileDisplayProperties.TextProperty) {
                    ShapefileDisplayProperties.TextProperty textProperty =
                        (ShapefileDisplayProperties.TextProperty) property;
                    if (textProperty.getVisibilityForText((String) value) == false)
                        continue;
                    shapeColor = textProperty.getColorForText((String) value);
                } else if (property instanceof ShapefileDisplayProperties.IntegerProperty) {
                    ShapefileDisplayProperties.IntegerProperty intProperty =
                        (ShapefileDisplayProperties.IntegerProperty) property;
                    int intValue = (int) ((Long) value).longValue();
                    if (intProperty.getVisibilityForValue(intValue) == false)
                        continue;
                    shapeColor = intProperty.getColorForValue(intValue);
                } else if (property instanceof ShapefileDisplayProperties.FloatingProperty) {
                    ShapefileDisplayProperties.FloatingProperty floatProperty =
                        (ShapefileDisplayProperties.FloatingProperty) property;
                    if (floatProperty.getVisibilityForValue((Double) value) == false)
                        continue;
                    shapeColor = floatProperty.getColorForValue((Double) value);
                }
            }
            graphics.setColor(properties.transform(shapeColor));

            // Draw the shape, filling it if it is an
            // ellipse (meaning a point is being drawn).
            Shape shape = drawable.getShape();
            ((Graphics2D) graphics).draw(shape);
            if (shape instanceof Ellipse2D.Float)
                ((Graphics2D) graphics).fill(shape);
        }

        // Unset the antialiasing in case it was set above.
        if (antiAliasingHint != null)
            ((Graphics2D) graphics).setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                                                     antiAliasingHint);



        /*
        Logger.logVerbose("Drawing of shapes for key " + key + " took " +
                          (System.currentTimeMillis() - startTime) + " milliseconds.");
        */




        // Determine the minimum path length required for
        // a path to be annotated.
        final double MINIMUM_PATH_LENGTH = MINIMUM_PATH_LENGTH_CELL_MULTIPLIER *
            annotationGrid.getCellWidth();

        // If drawables have not been fully cached, then
        // finish the job.
        if (drawablesDone == false) {

            // Iterate through the points that may be anno-
            // tated, recording which cell each one falls
            // within.
            for (int j = 0; j < annotatedPoints.size(); j++) {
                AnnotatedPoint point = annotatedPoints.get(j);
                int x = point.getPoint().x / cellWidth;
                int y = point.getPoint().y / cellHeight;
                if ((x < 0) || (x >= numberOfCellsHoriz) ||
                    (y < 0) || (y >= numberOfCellsVert))
                    continue;
                cachedPointsInCells[x][y].add(point);
            }

            // Iterate through all the different lists of
            // paths that were created above, keyed by text
            // strings, in order to determine which paths
            // are candidates for displaying their associa-
            // ted text.
            Iterator<String> iterator = pathsForNames.keySet().iterator();
            Rectangle displayRect = new Rectangle(width, height);
            while (iterator.hasNext()) {

                // For each path in the list, determine if
                // it is a candidate for displaying its
                // associated text.
                String text = iterator.next();
                LinkedList<ArrayList<Point>> paths = pathsForNames.get(text);
                for (int j = 0; j < paths.size(); j++) {

                    // Create a bounding rectangle from the
                    // list of points making up this path.
                    ArrayList<Point> points = paths.get(j);
                    Rectangle pathRect = null;
                    for (int k = 0; k < points.size(); k++) {
                        Point thisPoint = points.get(k);
                        if (k == 0)
                            pathRect = new Rectangle(thisPoint);
                        else
                            pathRect.add(thisPoint);
                    }

                    // Determine whether the visible part
                    // of the path's bounding rectangle has
                    // a large enough distance between dia-
                    // gonal bounding points to make this
                    // path a candidate for showing associ-
                    // ated text; if not, do nothing more
                    // with it.
                    pathRect = pathRect.intersection(displayRect);
                    double dist = -1.0;
                    if ((pathRect.width + pathRect.height < MINIMUM_PATH_LENGTH) &&
                        ((dist = Math.sqrt((double) ((pathRect.width * pathRect.width) +
                                                     (pathRect.height * pathRect.height)))) <
                         MINIMUM_PATH_LENGTH))
                        continue;

                    // Create a record for this path in-
                    // cluding its text and bounding box
                    // diagonal size.
                    if (dist < 0.0)
                        dist = Math.sqrt((double) ((pathRect.width * pathRect.width) +
                                                   (pathRect.height * pathRect.height)));
                    HashSet<ArrayList<AnnotatedPath>> cellsForPath =
                        new HashSet<ArrayList<AnnotatedPath>>();
                    AnnotatedPath annotatedPath = new AnnotatedPath(points, text,
                                                                    (int) (dist + 0.5),
                                                                    cellsForPath);



                    /*
                      long baseTime = System.currentTimeMillis();
                    */






                    // Place the record in the list associ-
                    // ated with each cell in which it is
                    // present, and a corresponding link to
                    // the cell from the path.
                    for (int y = 0; y < numberOfCellsVert; y++) {
                        for (int x = 0; x < numberOfCellsHoriz; x++) {
                            if (pathIntersectsRectangle(points, annotationGrid.
                                                        getCellBounds(x, y))) {
                                cachedPathsInCells[x][y].add(annotatedPath);
                                cellsForPath.add(cachedPathsInCells[x][y]);
                            }
                        }
                    }

                    /*
                      subTime += System.currentTimeMillis() - baseTime;
                    */
                }
            }

            // Iterate through the polygons for each text
            // string, determining which cells each one oc-
            // cupies, starting with the one with the lar-
            // gest bounding rectangle area for each text
            // string.  Also note which cell each polygon's
            // centroid falls within.
            Iterator<String> polyKeys = polygonsForName.keySet().iterator();
            while (polyKeys.hasNext()) {

                // Sort the polygons by bounding rectangle
                // area so that the largest one is checked
                // first, etc.
                ArrayList<AnnotatedPolygon> polygonList = polygonsForName.get(polyKeys.next());
                Collections.sort(polygonList);
                for (int j = 0; j < polygonList.size(); j++) {

                    // Determine which cell holds the cen-
                    // troid of the polygon, and record it.
                    AnnotatedPolygon polygon = polygonList.get(j);
                    int cellX = polygon.getCenterPoint().x / cellWidth;
                    int cellY = polygon.getCenterPoint().y / cellHeight;
                    if ((cellX >= 0) && (cellX < numberOfCellsHoriz) &&
                        (cellY >= 0) && (cellY < numberOfCellsVert))
                        cachedCentroidsInCells[cellX][cellY].add(polygon);

                    // Determine the cell boundaries for the
                    // rectangle that encompasses this poly-
                    // gon.
                    GeneralPath path = polygon.getPath();
                    Rectangle bounds = polygon.getBounds();
                    int minCellX = bounds.x / cellWidth;
                    if (minCellX < 0)
                        minCellX = 0;
                    else if (minCellX >= numberOfCellsHoriz)
                        minCellX = numberOfCellsHoriz - 1;
                    int maxCellX = (bounds.x + bounds.width) / cellWidth;
                    if (maxCellX < 0)
                        maxCellX = 0;
                    else if (maxCellX >= numberOfCellsHoriz)
                        maxCellX = numberOfCellsHoriz - 1;
                    int minCellY = bounds.y / cellHeight;
                    if (minCellY < 0)
                        minCellY = 0;
                    else if (minCellY >= numberOfCellsVert)
                        minCellY = numberOfCellsVert - 1;
                    int maxCellY = (bounds.y + bounds.height) / cellHeight;
                    if (maxCellY < 0)
                        maxCellY = 0;
                    else if (maxCellY >= numberOfCellsVert)
                        maxCellY = numberOfCellsVert - 1;

                    // Iterate through the cells within the
                    // bounds found above, determining for
                    // each one whether (assuming no other
                    // polygon has claimed ownership of the
                    // cell) the polygon fully encompasses
                    // the cell's area, and if so, marking
                    // the cell as the territory of this
                    // polygon.
                    boolean encompassesCell = false;
                    for (int x = minCellX; x <= maxCellX; x++) {
                        for (int y = minCellY; y <= maxCellY; y++) {
                            if (cachedPolygonsInCells[x][y] != null)
                                continue;
                            if (path.contains(annotationGrid.getCellBounds(x, y))) {
                                cachedPolygonsInCells[x][y] = polygon;
                                encompassesCell = true;
                            }
                        }
                    }

                    // If the polygon owns at least one
                    // cell unconditionally, then it may be
                    // able to be annotated; remember it
                    // for later.
                    if (encompassesCell)
                        cachedPolygons.add(polygon);
                }
            }
        }



        /*
        Logger.logVerbose("Potential candidate determination took " +
                          (System.currentTimeMillis() - startTime - subTime) +
                          " milliseconds.");
        Logger.logVerbose("Candidate visibility checking took " + subTime +
                          " milliseconds.");
        startTime = System.currentTimeMillis();
        */




        // If text placement has already been determined,
        // do nothing more.
        if (textPlacementDone)
            return null;

        // Create a flag for each cell indicating whether
        // or not that cell is already occupied by a text
        // annotation.
        boolean[][] cellsUsed = new boolean[numberOfCellsHoriz][numberOfCellsVert];

        // Set the font to be used for the labels.
        cachedFont = new Font(shapefileProperties.getFontName(),
                              (shapefileProperties.isFontBold() ? Font.BOLD : 0) +
                              (shapefileProperties.isFontItalic() ? Font.ITALIC : 0),
                              shapefileProperties.getFontSize());
        graphics.setFont(cachedFont);
        FontMetrics fm = graphics.getFontMetrics(cachedFont);

        // Create a hash table mapping polygon names to
        // flags indicating whether or not a polygon of
        // that particular name has been annotated al-
        // ready.
        HashMap<String, Boolean> polygonNamesAnnotated = new HashMap<String, Boolean>();

        // Iterate through the polygons that are to be
        // annotated if possible, attempting to anno-
        // tate each one once as close to its center
        // point as possible.
        for (int j = 0; j < cachedPolygons.size(); j++) {

            // Make sure that a polygon with this name
            // has not already been annotated, since
            // only one such polygon should be anno-
            // tated.
            AnnotatedPolygon polygon = cachedPolygons.get(j);
            if (polygonNamesAnnotated.get(polygon.getText()) != null)
                continue;

            // Get the polygon's center point, and find
            // the corresponding cell; if this cell is
            // outside the display bounds, adjust it
            // to be the closest cell on the display.
            boolean[] moreInDirection = { true, true, true, true };
            Point centerPoint = polygon.getCenterPoint();
            int centerX = centerPoint.x / cellWidth;
            if (centerX < 0)
                centerX = 0;
            if (centerX >= numberOfCellsHoriz)
                centerX = numberOfCellsHoriz - 1;
            int centerY = centerPoint.y / cellHeight;
            if (centerY < 0)
                centerY = 0;
            if (centerY >= numberOfCellsVert)
                centerY = numberOfCellsVert - 1;

            // Get the bounding rectangle of the polygon
            // (clipped to the display), and then deter-
            // mine what cells match those bounds, so as
            // to allow the iteration outward that is
            // done below to not pass beyond the bounds.
            Rectangle boundsRect = polygon.getBounds();
            int minX = boundsRect.x / cellWidth;
            if (minX < 0)
                minX = 0;
            else if (minX >= cellsUsed.length)
                minX = cellsUsed.length - 1;
            int minY = boundsRect.y / cellHeight;
            if (minY < 0)
                minY = 0;
            else if (minY >= cellsUsed[0].length)
                minY = cellsUsed[0].length - 1;
            int maxX = (boundsRect.x + boundsRect.width) / cellWidth;
            if (maxX < 0)
                maxX = 0;
            else if (maxX >= cellsUsed.length)
                maxX = cellsUsed.length - 1;
            int maxY = (boundsRect.y + boundsRect.height) / cellHeight;
            if (maxY < 0)
                maxY = 0;
            else if (maxY >= cellsUsed[0].length)
                maxY = cellsUsed[0].length - 1;

            // Iterate away from the center point's cell,
            // starting with that cell itself and then
            // moving outward in concentric rectangles
            // away from the center, attempting for each
            // cell through which the iteration passes to
            // place the label there.
            boolean notDone = true;
            for (int radius = 0, sidesFinished = 0; notDone && (sidesFinished < 4); radius++) {

                // Go clockwise around the center at the
                // current radius, starting at the top
                // left and proceeding to the top right,
                // then the bottom right, then to the
                // bottom left, and finally back to the
                // top left, checking each one in turn
                // to see if it is encompassed by this
                // polygon, and if so, attempt to place
                // the label there.  If the radius is 0,
                // just check the center cell itself.
                for (int direction = 0; notDone && (direction < (radius == 0 ? 1 : 4)); direction++) {

                    // If the iteration has already run
                    // off this edge of the display, do
                    // not bother with this side of the
                    // concentric rectangle.
                    if (moreInDirection[direction] == false)
                        continue;

                    // Determine how many cells should
                    // be iterated through at this
                    // radius.  At radius 0, only the
                    // center cell itself should be
                    // checked.
                    int total = (radius * 2) + (radius == 0 ? 1 : 0);

                    // Determine the starting X cell,
                    // adjusting the total number of
                    // cells that must be processed at
                    // this radius for this side of
                    // the concentric rectangle if the
                    // starting X starts outside the
                    // polygon bounds, and the starting
                    // X itself.
                    int startX = centerX + (direction % 3 == 0 ? -radius : radius);
                    if (startX < minX) {
                        if (moreInDirection[LEFT]) {
                            moreInDirection[LEFT] = false;
                            sidesFinished++;
                        }
                        if (direction == ABOVE) {
                            total += startX - minX;
                            startX = minX;
                        } else
                            continue;
                    }
                    if (startX > maxX) {
                        if (moreInDirection[RIGHT]) {
                            moreInDirection[RIGHT] = false;
                            sidesFinished++;
                        }
                        if (direction == BELOW) {
                            total -= startX - maxX;
                            startX = maxX;
                        } else                            
                            continue;
                    }

                    // Determine the starting Y cell,
                    // adjusting the total number of
                    // cells that must be processed at
                    // this radius for this side of
                    // the concentric rectangle if the
                    // starting Y starts outside the
                    // polygon bounds, and the starting
                    // Y itself.
                    int startY = centerY + (direction < 2 ? -radius : radius);
                    if (startY < minY) {
                        if (moreInDirection[ABOVE]) {
                            moreInDirection[ABOVE] = false;
                            sidesFinished++;
                        }
                        if (direction == RIGHT) {
                            total += startY - minY;
                            startY = minY;
                        } else
                            continue;
                    }
                    if (startY > maxY) {
                        if (moreInDirection[BELOW]) {
                            moreInDirection[BELOW] = false;
                            sidesFinished++;
                        }
                        if (direction == LEFT) {
                            total -= startY - maxY;
                            startY = maxY;
                        } else                            
                            continue;
                    }

                    // Iterate through the cells to be
                    // checked on this side of the
                    // concentric rectangle, checking
                    // each.
                    for (int count = 0; notDone && (count < total); count++) {

                        // Determine the coordinates of
                        // the cell being checked; if it
                        // is outside the display bounds,
                        // do nothing more with this
                        // side of the concentric rec-
                        // tangle.
                        int x = startX + (direction % 2 == 1 ? 0 : count *
                                          (direction == ABOVE ? 1 : -1));
                        int y = startY + (direction % 2 == 0 ? 0 : count *
                                          (direction == RIGHT ? 1 : -1));
                        if ((x < minX) || (x > maxX) || (y < minY) || (y > maxY))
                            break;

                        // Do nothing with this cell if
                        // it is already occupied or
                        // does not belong to this
                        // polygon.
                        if (cellsUsed[x][y] || (cachedPolygonsInCells[x][y] != polygon))
                            continue;

                        // Get the basic text rectangle.
                        Rectangle boundingRect =
                            getBasicBoundsOfText(polygon.getText(), fm);
                        Point textCenter =
                            (radius == 0 ? centerPoint :
                             new Point((x * cellWidth) + (cellWidth / 2),
                                       (y * cellHeight) + (cellHeight / 2)));
                        boundingRect.x += textCenter.x;
                        boundingRect.y += textCenter.y;

                        // Determine what cells would be
                        // occupied by this bounding
                        // rectangle; if it occupies
                        // cells that are already used,
                        // do nothing more with this
                        // polygon in the context of
                        // this cell.  This will also
                        // ensure that the rectangle
                        // only occupies cells that are
                        // encompassed by this polygon.
                        // If it returns true, then all
                        // the cells for this rectangle
                        // are now marked as used.
                        Point textCell = new Point(x, y);
                        HashSet<Point> cellsForLabel =
                            canAxesAlignedRectangleBeDrawn(boundingRect, cellWidth,
                                                           cellHeight, cellsUsed,
                                                           textCell, cachedPolygonsInCells,
                                                           polygon);
                        if (cellsForLabel == null)
                            continue;

                        // Mark the name of this polygon
                        // as having already been used
                        // for an annotation.
                        polygonNamesAnnotated.put(polygon.getText(), true);

                        // Cache the label for painting
                        // later.
                        cachedHiPriLabelStrings.add(polygon.getText());
                        cachedHiPriLabelAngles.add(0.0);
                        Point startPoint = new Point(boundingRect.x,
                                                     boundingRect.y + fm.getMaxAscent());
                        cachedHiPriLabelPoints.add(startPoint);
                        cachedHiPriLabelCells.add(cellsForLabel);
                        cachedHiPriLabelVisualEmphasisPoints.add(null);

                        // Add the cell holding the cen-
                        // ter of the text to the list
                        // of cells previously used to
                        // annotate this polygon.
                        polygon.addToAnnotatedCells(textCell);

                        // Mark this polygon as done.
                        notDone = false;
                    }
                }
            }
        }

        // Iterate through the cells, placing a text
        // label in each free one that has a point,
        // polygon centroid, or path that is a candidate
        // for this treatment present within it.  The
        // cells are checked line by line, starting with
        // the center line, then alternating between the
        // ones above and those below; similarly, each
        // line is checked cell by cell, starting with
        // the center one, then alternating between the
        // cells to the left and to the right.
        boolean moreBelow = true, moreAbove = true;
        int startY = numberOfCellsVert / 2;
        for (int iterCountY = 0, y = startY; moreBelow || moreAbove; iterCountY++,
                 y = startY + (((iterCountY + 1) / 2) * (iterCountY % 2 == 0 ? -1 : 1))) {
//                 y = startY + (((iterCountY + 1) / 2) * (iterCountY % 2 == 0 ? -8 : 8))) {
            if (y < 0) {
                moreBelow = false;
                continue;
            } else if (y >= numberOfCellsVert) {
                moreAbove = false;
                continue;
            }
            boolean moreLeft = true, moreRight = true;
            int startX = numberOfCellsHoriz / 2;
            for (int iterCountX = 0, x = startX; moreLeft || moreRight; iterCountX++,
                     x = startX + (((iterCountX + 1) / 2) * (iterCountX % 2 == 0 ? -1 : 1))) {
//                     x = startX + (((iterCountX + 1) / 2) * (iterCountX % 2 == 0 ? -8 : 8))) {
                if (x < 0) {
                    moreLeft = false;
                    continue;
                } else if (x >= numberOfCellsHoriz) {
                    moreRight = false;
                    continue;
                }

                // If the cell is already occupied by a
                // label, or it has no candidate shapes
                // within it, do nothing with it.
                if (cellsUsed[x][y] || (cachedPointsInCells[x][y].size() +
                                        cachedCentroidsInCells[x][y].size() +
                                        cachedPathsInCells[x][y].size() == 0))
                    continue;

                // Determine the center point of this
                // cell.
                Rectangle cellRect = annotationGrid.getCellBounds(x, y);
                Point cellCenter = new Point(cellRect.x + (cellRect.width / 2),
                                             cellRect.y + (cellRect.height / 2));

                // Iterate through the list of candi-
                // date points, trying for each one to
                // draw its label.  Once one is drawn,
                // the iteration is stopped, as this
                // cell is now considered occupied.
                for (int k = 0; (cellsUsed[x][y] == false) &&
                         (k < cachedPointsInCells[x][y].size()); k++) {
                    AnnotatedPoint point = cachedPointsInCells[x][y].get(k);

                    // Get the basic text rectangle.
                    Rectangle boundingRect = getBasicBoundsOfText(point.getText(), fm);

                    // Find the point just above the
                    // point where the text will be
                    // centered.
                    Point textCenter = new Point(point.getPoint().x,
                                                 point.getPoint().y -
                                                 ((boundingRect.height / 2) +
                                                  POINT_ANNOTATION_VERTICAL_OFFSET));

                    // Determine what cell the text
                    // center is in.
                    Point textCell = new Point(textCenter.x / cellWidth,
                                               textCenter.y / cellHeight);

                    // If this cell is already occupied,
                    // do nothing more with this point
                    // in the context of the original
                    // cell.  Otherwise, remember it so
                    // that it can be marked as occupied
                    // later if this label gets drawn.
                    if ((textCell.x >= 0) && (textCell.x < cellsUsed.length) &&
                        (textCell.y >= 0) && (textCell.y < cellsUsed[0].length)) {
                        if (cellsUsed[textCell.x][textCell.y])
                            continue;
                    }

                    // Adjust the bounding rectangle to
                    // be centered on the text center-
                    // point, then rotate it.
                    boundingRect.x += textCenter.x;
                    boundingRect.y += textCenter.y;
                        
                    // Determine what cells would be occu-
                    // pied by this bounding rectangle; if
                    // it occupies cells that are already
                    // used, do nothing more with this
                    // point in the context of this cell.
                    // Otherwise, the cells will be marked
                    // as used when this method returns.
                    HashSet<Point> cellsForLabel =
                        canAxesAlignedRectangleBeDrawn(boundingRect, cellWidth,
                                                       cellHeight, cellsUsed, textCell,
                                                       null, null);
                    if (cellsForLabel == null)
                        continue;

                    // Cache the label information for
                    // painting later.
                    cachedHiPriLabelStrings.add(point.getText());
                    cachedHiPriLabelAngles.add(0.0);
                    Point startPoint = new Point(boundingRect.x,
                                                 boundingRect.y + fm.getMaxAscent());
                    cachedHiPriLabelPoints.add(startPoint);
                    cachedHiPriLabelCells.add(cellsForLabel);
                    cachedHiPriLabelVisualEmphasisPoints.add(point.getPoint());
                }

                // Iterate through the list of candi-
                // date polygons that have centroids
                // within this cell, trying for each
                // one to draw its label.  Once one
                // is drawn, the iteration is stopped,
                // as this cell is now considered
                // occupied.
                for (int k = 0; (cellsUsed[x][y] == false) &&
                         (k < cachedCentroidsInCells[x][y].size()); k++) {

                    // Make sure that a polygon with
                    // this name has not already been
                    // annotated, since only one such
                    // polygon should be annotated.
                    AnnotatedPolygon polygon = cachedCentroidsInCells[x][y].get(k);
                    if (polygonNamesAnnotated.get(polygon.getText()) != null)
                        continue;

                    // Get the basic text rectangle.
                    Rectangle boundingRect = getBasicBoundsOfText(polygon.getText(), fm);

                    // Find the point just above the
                    // centroid where the text will
                    // be centered.
                    Point textCenter = new Point(polygon.getCenterPoint().x,
                                                 polygon.getCenterPoint().y -
                                                 ((boundingRect.height / 2) +
                                                  POINT_ANNOTATION_VERTICAL_OFFSET));

                    // Determine what cell the text
                    // center is in.
                    Point textCell = new Point(textCenter.x / cellWidth,
                                               textCenter.y / cellHeight);

                    // If this cell is already occupied,
                    // do nothing more with this polygon
                    // in the context of the original
                    // cell.  Otherwise, remember it so
                    // that it can be marked as occupied
                    // later if this label gets drawn.
                    if ((textCell.x >= 0) && (textCell.x < cellsUsed.length) &&
                        (textCell.y >= 0) && (textCell.y < cellsUsed[0].length)) {
                        if (cellsUsed[textCell.x][textCell.y])
                            continue;
                    }

                    // Adjust the bounding rectangle to
                    // be centered on the text center-
                    // point, then rotate it.
                    boundingRect.x += textCenter.x;
                    boundingRect.y += textCenter.y;
                        
                    // Determine what cells would be occu-
                    // pied by this bounding rectangle; if
                    // it occupies cells that are already
                    // used, do nothing more with this
                    // polygon in the context of this cell.
                    // Otherwise, the cells will be marked
                    // as used when this method returns.
                    HashSet<Point> cellsForLabel =
                        canAxesAlignedRectangleBeDrawn(boundingRect, cellWidth,
                                                       cellHeight, cellsUsed, textCell,
                                                       null, null);
                    if (cellsForLabel == null)
                        continue;

                    // Mark the name of this polygon as
                    // having already been used for an
                    // annotation.
                    polygonNamesAnnotated.put(polygon.getText(), true);

                    // Cache the label information for
                    // painting later.
                    cachedHiPriLabelStrings.add(polygon.getText());
                    cachedHiPriLabelAngles.add(0.0);
                    Point startPoint = new Point(boundingRect.x,
                                                 boundingRect.y + fm.getMaxAscent());
                    cachedHiPriLabelPoints.add(startPoint);
                    cachedHiPriLabelCells.add(cellsForLabel);
                    cachedHiPriLabelVisualEmphasisPoints.add(null);
                }

                // Iterate through the list of candi-
                // date paths, trying for each one to
                // draw its label.  Once one is drawn,
                // the iteration is stopped, as this
                // cell is now considered occupied.
                // Sort the list first to ensure that
                // those paths that have highest pri-
                // ority are checked first.
                Collections.sort(cachedPathsInCells[x][y]);
                for (int k = 0; (cellsUsed[x][y] == false) &&
                         (k < cachedPathsInCells[x][y].size()); k++) {
                    AnnotatedPath path = cachedPathsInCells[x][y].get(k);

                    // Get the basic text rectangle.
                    Rectangle boundingRect = getBasicBoundsOfText(path.getText(), fm);

                    // Find the point along the line
                    // that is the closest to the line,
                    // and the angle of the line at
                    // that point; if the angle is less
                    // than 0, then no point was found,
                    // so do nothing more with this
                    // path in this cell.
                    Point textCenter = new Point();
                    double angle =
                        findNearestPointAndAngleOnPath(cellCenter, path.getPath(),
                                                       boundingRect.width, textCenter);
                    if (angle < 0.0)
                        continue;

                    // Determine what cell the text
                    // center is in.
                    Point textCell = new Point(textCenter.x / cellWidth,
                                               textCenter.y / cellHeight);

                    // If this cell is already occupied,
                    // do nothing more with this path
                    // in the context of the original
                    // cell.
                    if ((textCell.x >= 0) && (textCell.x < cellsUsed.length) &&
                        (textCell.y >= 0) && (textCell.y < cellsUsed[0].length) &&
                        cellsUsed[textCell.x][textCell.y])
                        continue;

                    // If the text center is too close
                    // to previous annotations of this
                    // path, skip it.
                    ArrayList<Point> annotatedCells = path.getAnnotatedCells();
                    int cellsTakenByAnnotationLength =
                        boundingRect.width / annotationGrid.getCellWidth();
                    boolean tooClose = false;
                    for (int m = 0; m < annotatedCells.size(); m++) {
                        Point cell = annotatedCells.get(m);
                        if (Math.abs(cell.x - textCell.x) + Math.abs(cell.y - textCell.y) <
                            cellsTakenByAnnotationLength + ANNOTATION_SPACING_FOR_SAME_PATH) {
                            tooClose = true;
                            break;
                        }
                    }
                    if (tooClose)
                        continue;
                    



                    /*
                    // GREEN LINE FOR ROAD BEING
                    // ANNOTATED.
                    graphics.setColor(Color.green);
                    int[] xs = new int[path.getPath().size()];
                    int[] ys = new int[path.getPath().size()];
                    for (int m = 0; m < xs.length; m++) {
                        xs[m] = path.getPath().get(m).x;
                        ys[m] = path.getPath().get(m).y;
                    }
                    graphics.drawPolyline(xs, ys, xs.length);
                    */

                    /*

                    // TEXT UNROTATED AND AT CENTER
                    // OF CELL
                    graphics.setColor(properties.transform(Color.white));
                    graphics.drawString(path.getText(), cellCenter.x - (boundingRect.width / 2),
                                        cellCenter.y + (boundingRect.height / 2) - fm.getMaxDescent());
                    break;
                    */






                    // Adjust the bounding rectangle to
                    // be centered on the text center-
                    // point, then rotate it.
                    boundingRect.x += textCenter.x;
                    boundingRect.y += textCenter.y;
                    AffineTransform at = new AffineTransform();
                    at.rotate(Math.toRadians(angle), textCenter.x, textCenter.y);
                        
                    // Determine what cells would be occu-
                    // pied by this rotated rectangle; if
                    // it occupies cells that are already
                    // used, do nothing more with this
                    // path in the context of this cell.
                    // Otherwise, the cells will be marked
                    // as used when this method returns.
                    HashSet<Point> cellsForLabel =
                        canArbitraryRectangleBeDrawn(at.createTransformedShape(boundingRect).
                                                     getPathIterator(null), cellWidth,
                                                     cellHeight, cellsUsed, textCell);
                    if (cellsForLabel == null)
                        continue;

                    // Add the cell holding the center
                    // of the text to the list of cells
                    // previously used to annotate this
                    // path.
                    path.addToAnnotatedCells(textCell);

                    // Subtract the length of the text
                    // from the length associated with
                    // this path's record.
                    int priority = path.getPriority();
                    path.setPriority(priority - boundingRect.width);





                    /*
                    // RED RECTANGLE AROUND TARGET CELLS
                    cellIterator = cellsToBeUsed.iterator();
                    while (cellIterator.hasNext()) {
                        Point cell = cellIterator.next();
                        graphics.setColor(Color.red);
                        graphics.drawRect(cells[cell.x][cell.y].x, cells[cell.x][cell.y].y,
                                          cells[cell.x][cell.y].width, cells[cell.x][cell.y].height);
                    }
                    */

                    /*
                    // YELLOW RECTANGLE AROUND VERTEX
                    // CELLS, MAGENTA AROUND TEXT CENTER CELL.
                    cellIterator = cellsToBeUsed2.iterator();
                    while (cellIterator.hasNext()) {
                        Point cell = cellIterator.next();
                        graphics.setColor(Color.yellow);
                        graphics.drawRect(cells[cell.x][cell.y].x + 1, cells[cell.x][cell.y].y + 1,
                                          cells[cell.x][cell.y].width - 2, cells[cell.x][cell.y].height - 2);
                    }
                    if ((textCell.x >= 0) && (textCell.x < cells.length) &&
                        (textCell.y >= 0) && (textCell.y < cells[0].length)) {
                        graphics.setColor(Color.magenta);
                        graphics.drawRect(cells[textCell.x][textCell.y].x + 1, cells[textCell.x][textCell.y].y + 1,
                                          cells[textCell.x][textCell.y].width - 2, cells[textCell.x][textCell.y].height - 2);
                        graphics.drawRect(cells[textCell.x][textCell.y].x + 2, cells[textCell.x][textCell.y].y + 2,
                                          cells[textCell.x][textCell.y].width - 4, cells[textCell.x][textCell.y].height - 4);
                    }
                    */







                    // Rotate the max ascent for the font
                    // by the angle of the text; this
                    // yields the X and Y offsets needed
                    // to make up for the fact that the
                    // text is drawn by the graphic object
                    // with the point at the font's base-
                    // line.
                    double adjustedAngle = Math.toRadians(angle);
                    double xDelta = -fm.getMaxAscent() * Math.sin(adjustedAngle);
                    double yDelta = fm.getMaxAscent() * Math.cos(adjustedAngle);
                    Dimension ascentOffsets =
                        new Dimension((int) (xDelta + (xDelta < 0.0 ? -0.5 : 0.5)),
                                      (int) (yDelta + (yDelta < 0.0 ? -0.5 : 0.5)));

                    // Rotate the text's upper left cor-
                    // ner around the centerpoint of
                    // this annotation to arrive at the
                    // starting point for text drawing.
                    xDelta = ((boundingRect.x - textCenter.x) * Math.cos(adjustedAngle)) -
                        ((boundingRect.y - textCenter.y) * Math.sin(adjustedAngle));
                    yDelta = ((boundingRect.x - textCenter.x) * Math.sin(adjustedAngle)) +
                        ((boundingRect.y - textCenter.y) * Math.cos(adjustedAngle));
                    Point startPoint = new Point(textCenter.x + ascentOffsets.width +
                                                 (int) (xDelta + (xDelta < 0.0 ? -0.5 : 0.5)),
                                                 textCenter.y + ascentOffsets.height +
                                                 (int) (yDelta + (yDelta < 0.0 ? -0.5 : 0.5)));

                    // Cache the label for painting later.
                    if ((path.getAnnotatedCells().size() < 2) &&
                        (priority >= MINIMUM_PATH_LENGTH * 2)) {
                        cachedHiPriLabelStrings.add(path.getText());
                        cachedHiPriLabelAngles.add(adjustedAngle);
                        cachedHiPriLabelPoints.add(startPoint);
                        cachedHiPriLabelCells.add(cellsForLabel);
                        cachedHiPriLabelVisualEmphasisPoints.add(null);
                    } else {
                        cachedOtherLabelStrings.add(path.getText());
                        cachedOtherLabelAngles.add(adjustedAngle);
                        cachedOtherLabelPoints.add(startPoint);
                        cachedOtherLabelCells.add(cellsForLabel);
                    }
                }
            }
        }

        // Iterate through the cells just as was done
        // above, placing a text label in each free one
        // that is encompassed by a polygon that does
        // not have an already-drawn label too close to
        // this cell.  As above, the cells are checked
        // line by line, starting with the center line,
        // then alternating between the ones above and
        // those below; similarly, each line is checked
        // cell by cell, starting with the center one,
        // then alternating between the cells to the
        // left and to the right.
        moreBelow = moreAbove = true;
        startY = numberOfCellsVert / 2;
        for (int iterCountY = 0, y = startY; moreBelow || moreAbove; iterCountY++,
                 y = startY + (((iterCountY + 1) / 2) * (iterCountY % 2 == 0 ? -1 : 1))) {
            if (y < 0) {
                moreBelow = false;
                continue;
            } else if (y >= numberOfCellsVert) {
                moreAbove = false;
                continue;
            }
            boolean moreLeft = true, moreRight = true;
            int startX = numberOfCellsHoriz / 2;
            for (int iterCountX = 0, x = startX; moreLeft || moreRight; iterCountX++,
                     x = startX + (((iterCountX + 1) / 2) * (iterCountX % 2 == 0 ? -1 : 1))) {
                if (x < 0) {
                    moreLeft = false;
                    continue;
                } else if (x >= numberOfCellsHoriz) {
                    moreRight = false;
                    continue;
                }

                // If the cell is already occupied by a
                // label, or it is not encompassed by
                // a polygon, or if the polygon that is
                // occupying it has a name that has al-
                // ready been used once for a polygon's
                // annotation, do nothing with it.
                if (cellsUsed[x][y] || (cachedPolygonsInCells[x][y] == null))
                    continue;
                AnnotatedPolygon polygon = cachedPolygonsInCells[x][y];
                if (polygonNamesAnnotated.get(polygon.getText()) != null)
                    continue;

                // Get the basic text rectangle.
                Rectangle boundingRect = getBasicBoundsOfText(polygon.getText(), fm);

                // If the cell is too close to previous
                // annotations of this polygon, skip it.
                ArrayList<Point> annotatedCells = polygon.getAnnotatedCells();
                int cellsTakenByAnnotationLength =
                    boundingRect.width / annotationGrid.getCellWidth();
                boolean tooClose = false;
                for (int m = 0; m < annotatedCells.size(); m++) {
                    Point cell = annotatedCells.get(m);
                    if (Math.abs(cell.x - x) + Math.abs(cell.y - y) <
                        cellsTakenByAnnotationLength +
                        ANNOTATION_SPACING_FOR_SAME_POLYGON) {
                        tooClose = true;
                        break;
                    }
                }
                if (tooClose)
                    continue;

                // Adjust the bounding rectangle to be
                // centered on the center of the cell,
                // then rotate it.
                boundingRect.x += (x * cellWidth) + (cellWidth / 2);
                boundingRect.y += (y * cellHeight) + (cellHeight / 2);
                        
                // Determine what cells would be occu-
                // pied by this bounding rectangle; if
                // it occupies cells that are already
                // used, do nothing more with this
                // point in the context of this cell.
                // Otherwise, the cells will be marked
                // as used when this method returns.
                Point textCell = new Point(x, y);
                HashSet<Point> cellsForLabel =
                    canAxesAlignedRectangleBeDrawn(boundingRect, cellWidth,
                                                   cellHeight, cellsUsed, textCell,
                                                   null, null);
                if (cellsForLabel == null)
                    continue;

                // Mark the name of this polygon as
                // having already been used for an
                // annotation.
                polygonNamesAnnotated.put(polygon.getText(), true);

                // Cache the label for painting later.
                Point startPoint = new Point(boundingRect.x,
                                             boundingRect.y + fm.getMaxAscent());
                if (polygon.getAnnotatedCells().size() == 0) {
                    cachedHiPriLabelStrings.add(polygon.getText());
                    cachedHiPriLabelAngles.add(0.0);
                    cachedHiPriLabelPoints.add(startPoint);
                    cachedHiPriLabelCells.add(cellsForLabel);
                    cachedHiPriLabelVisualEmphasisPoints.add(null);
                } else {
                    cachedOtherLabelStrings.add(polygon.getText());
                    cachedOtherLabelAngles.add(0.0);
                    cachedOtherLabelPoints.add(startPoint);
                    cachedOtherLabelCells.add(cellsForLabel);
                }

                // Add the cell holding the center of
                // the text to the list of cells pre-
                // viously used to annotate the poly-
                // gon.
                polygon.addToAnnotatedCells(textCell);
            }
        }

        /*
        Logger.logVerbose("Annotation calculating took " + (System.currentTimeMillis() - startTime) + " milliseconds.");
        */

        // Always return null, as no waiting for an image
        // needs to be done.
        return null;
    }

    /**
     * Paint any high-priority annotations, i.e. those
     * labels that the depictable rates as most important
     * to display.
     *
     * @param  graphics       Graphics object to be used.
     * @param  visualInfo     Visual information about
     *                        this map or product, if any.
     * @param  depictProps    Display properties to be
     *                        used, if any.
     * @param  observer       Image observer that is to be
     *                        notified if a paint cannot
     *                        be completed because an
     *                        image has not yet finished
     *                        loading.
     * @param  annotationGrid Annotation grid indicating
     *                        which cells are occupied
     *                        and which are not. This
     *                        grid is to be updated with
     *                        any cells into which this
     *                        method paints annotations
     *                        marked as used when this
     *                        method completes execution.
     * @return If painting could not be completed because
     *         a required image has not yet finished
     *         loading, the image that is still loading
     *         is returned.  Otherwise, <code>null</code>
     *         is returned.
     */
    public Image paintPrimaryAnnotations(Graphics graphics,
                                         ProductVisualInfo visualInfo,
                                         DepictableDisplayProperties depictProps,
                                         ImageObserver observer,
                                         AnnotationGrid annotationGrid) {
        return paintAnnotations(true, graphics, visualInfo, depictProps, observer,
                                annotationGrid);
    }

    /**
     * Paint all other (non-high-priority) annotations,
     * i.e. those labels that the depictable rates as less
     * important to display, but still worth displaying.
     *
     * @param  graphics       Graphics object to be used.
     * @param  visualInfo     Visual information about
     *                        this map or product, if any.
     * @param  depictProps    Display properties to be
     *                        used, if any.
     * @param  observer       Image observer that is to be
     *                        notified if a paint cannot
     *                        be completed because an
     *                        image has not yet finished
     *                        loading.
     * @param  annotationGrid Annotation grid indicating
     *                        which cells are occupied
     *                        and which are not. This
     *                        grid is to be updated with
     *                        any cells into which this
     *                        method paints annotations
     *                        marked as used when this
     *                        method completes execution.
     * @return If painting could not be completed because
     *         a required image has not yet finished
     *         loading, the image that is still loading
     *         is returned.  Otherwise, <code>null</code>
     *         is returned.
     */
    public Image paintOtherAnnotations(Graphics graphics,
                                       ProductVisualInfo visualInfo,
                                       DepictableDisplayProperties depictProps,
                                       ImageObserver observer,
                                       AnnotationGrid annotationGrid) {
        return paintAnnotations(false, graphics, visualInfo, depictProps, observer,
                                annotationGrid);
    }


    // Protected Methods

    /**
     * Read the specified number of bytes from the given
     * stream and return a stream that allows access to
     * the read bytes.
     *
     * @param  inStream Stream from which to read the
     *                  bytes.
     * @param  numBytes Number of bytes to read.
     * @return Stream allowing access to the read bytes.
     * @throws Exception if an error occurs.
     */
    protected ByteArrayInputStream getByteStream(InputStream inStream,
                                                 int numBytes) throws Exception {
        byte[] buffer = new byte[numBytes];
        for (int read = 0, justRead = 0; read < numBytes; read += justRead) {
            justRead = inStream.read(buffer, read, (numBytes - read));
            if (justRead == -1)
                throw new Exception("premature EOF while reading into byte stream");
        }
        return new ByteArrayInputStream(buffer);
    }

    /**
     * Read the SHP file from the specified input stream.
     *
     * @param  inStream Input stream from which to read
     *                  the SHP file data.
     * @param  size     Size of the SHP file.
     * @throws Exception if an error occurs.
     */
    protected void readSHP(InputStream inStream, long size) throws Exception {

        // Read in the header of the shapefile (first 100
        // bytes), checking the file code and length to
        // ensure they are valid.
        byte[] intBytes = new byte[4];
        byte[] doubleBytes = new byte[8];
        int code = 0;
        if (size < 100)
            throw new Exception("invalid SHP file: premature EOF while reading header");
        try {
            code = getBigEndianIntegerFromStream(inStream, intBytes);
        } catch (Exception e) {
            throw new Exception("invalid SHP file: premature EOF while reading header");
        }
        if (code != 9994)
            throw new Exception("invalid SHP file: does not begin with correct code");
        long length = 0;
        try {
            skipBytesInStream(inStream, 20);
            length = getBigEndianIntegerFromStream(inStream, intBytes) * 2;
        } catch (Exception e) {
            throw new Exception("invalid SHP file: premature EOF while reading header");
        }
        if (length != size)
            throw new Exception("invalid SHP file: recorded file length of " + length +
                                " bytes does not match actual file length of " + size +
                                " bytes.");
        skipBytesInStream(inStream, 72);

        // Read in each record.
        int readCount = 100;
        boolean foundUnhandledType = false;
        for (int recordCount = 0; readCount < length; recordCount++) {

            // Read in the content length of the record,
            // not including the eight bytes that head
            // up that record (holding the identifier
            // and the content length itself).
            skipBytesInStream(inStream, 4);
            int recordLength = getBigEndianIntegerFromStream(inStream, intBytes) * 2;
            readCount += 8;

            // Read in the type of the shape within this
            // record.
            int type = getLittleEndianIntegerFromStream(inStream, intBytes);

            // Read the rest of the record differently
            // depending upon its type. Only point,
            // polyline, and polygon types are handled.
            if (type == POINT_TYPE)
                shapes.add(new GeoPoint(recordCount, null,
                                        getLatLongFromStream(inStream, doubleBytes)));
            else if (type == POLYLINE_TYPE) {

                // Skip the bounding box, then get the
                // number of parts and the number of
                // points.
                skipBytesInStream(inStream, 32);
                int numParts = getLittleEndianIntegerFromStream(inStream, intBytes);
                int numPoints = getLittleEndianIntegerFromStream(inStream, intBytes);

                // Read in the starting indices of the
                // parts in the points list, adding the
                // number of points as an extra starting
                // index to make the division of points
                // into parts easier below.
                int[] partIndices = new int[numParts + 1];
                for (int j = 0; j < numParts; j++)
                    partIndices[j] = getLittleEndianIntegerFromStream(inStream, intBytes);
                partIndices[partIndices.length - 1] = numPoints;

                // Read in the list of parts, each being
                // an array of points for that part, and
                // join together any part that is found
                // to share an endpoint with any other
                // part from this polyline.
                ArrayList<ArrayList<LatLong>> parts = new ArrayList<ArrayList<LatLong>>();
                int pointCount = 0;
                for (int j = 0; (j < numParts) && (pointCount < numPoints); j++) {

                    // Read in the points for this part.
                    ArrayList<LatLong> points = new ArrayList<LatLong>();
                    for (int k = 0; k < partIndices[j + 1] - partIndices[j]; k++,
                             pointCount++)
                        points.add(getLatLongFromStream(inStream, doubleBytes));

                    // Try joining this part at its
                    // starting point with other parts
                    // first, then try the same thing
                    // with the part's ending point.
                    for (int end = 0; end < 2; end++) {

                        // Iterate through the parts,
                        // attempting to join the chosen
                        // end of this part with either
                        // end of the other parts.
                        LatLong thisPoint = points.get(end != 0 ? points.size() - 1 : 0);
                        int joinWith = parts.size();
                        boolean joinAtStart = false;
                        for (int m = 0; m < parts.size(); m++) {

                            // See if this part joins
                            // with either end of the
                            // other part.
                            LatLong point1 = parts.get(m).get(0);
                            LatLong point2 = parts.get(m).get(parts.get(m).size() - 1);
                            for (int n = 0; n < 2; n++) {

                                // If the points are
                                // the same, the part
                                // should be joined
                                // with the other
                                // part at the appro-
                                // priate end.
                                if ((n == 1 ? point2 : point1).equals(thisPoint)) {
                                    joinWith = m;
                                    joinAtStart = (n == 0);
                                    break;
                                }
                            }
                            if (joinWith < parts.size())
                                break;
                        }

                        // If joining should occur, do
                        // it now.  First reverse the
                        // other part's points if the
                        // join requires it; then pre-
                        // pend or append it to this
                        // part.  Also remove the other
                        // part, as it is no longer a
                        // separate entity.
                        if (joinWith < parts.size()) {
                            ArrayList<LatLong> otherPoints = parts.remove(joinWith);
                            if ((joinAtStart && (end == 0)) ||
                                (joinAtStart == false) && (end != 0))
                                Collections.reverse(otherPoints);
                            points.addAll((end != 0 ? points.size() : 0), otherPoints);
                        }
                    }

                    // Add the compiled part to the
                    // list of parts.
                    parts.add(points);
                }

                // Add each part as a separate shape,
                // but with the same identifier to
                // ensure that they all receive the
                // same property values.
                for (int j = 0; j < parts.size(); j++) {
                    LatLong[] points = new LatLong[parts.get(j).size()];
                    shapes.add(new GeoPath(recordCount, null,
                                           parts.get(j).toArray(points)));
                }
            } else if (type == POLYGON_TYPE) {

                // Skip the bounding box, then get the
                // number of parts and the number of
                // points.
                skipBytesInStream(inStream, 32);
                int numParts = getLittleEndianIntegerFromStream(inStream, intBytes);
                int numPoints = getLittleEndianIntegerFromStream(inStream, intBytes);

                // Read in the starting indices of the
                // parts in the points list, adding the
                // number of points as an extra starting
                // index to make the division of points
                // into parts easier below.
                int[] partIndices = new int[numParts + 1];
                for (int j = 0; j < numParts; j++)
                    partIndices[j] = getLittleEndianIntegerFromStream(inStream, intBytes);
                partIndices[partIndices.length - 1] = numPoints;

                // Read in the list of parts, each being
                // an array of points for that part.
                // Since the shapefile spec requires that
                // the last point in each part be the
                // same as the first, the last point can
                // be ignored for the depictable's pur-
                // poses.
                LatLong[][] parts = new LatLong[numParts][];
                int pointCount = 0;
                for (int j = 0; (j < numParts) && (pointCount < numPoints); j++) {

                    // Read in the points for this part.
                    parts[j] = new LatLong[partIndices[j + 1] - (partIndices[j] + 1)];
                    for (int k = 0; k < partIndices[j + 1] - (partIndices[j] + 1); k++,
                             pointCount++)
                        parts[j][k] = getLatLongFromStream(inStream, doubleBytes);
                    skipBytesInStream(inStream, 16);
                }

                // Add the polygon.
                shapes.add(new GeoPolygon(recordCount, null, parts));
            } else if (foundUnhandledType == false) {
                skipBytesInStream(inStream, recordLength);
                foundUnhandledType = true;
                Logger.logProblem("Encountered one or more unhandled shapefile types; " +
                                  "ignoring all such shapes.");
            }
            readCount += recordLength;
        }
    }

	/**
	 * Skip the specified number of bytes in the input
     * stream.
	 *
     * @param  inStream  Input stream in which to skip
     *                   the bytes.
     * @param  numToSkip Number of bytes to skip.
     * @throws Exception if an error occurs.
	 */
    protected void skipBytesInStream(InputStream inStream, int numToSkip) throws Exception {

        // Attempt to skip the bytes all at once, but
        // skip any individually that do not get skipped
        // the first time around (since the stream may
        // not have the bytes available to skip the first
        // time around).
        int num = (int) inStream.skip(numToSkip);
        while (num++ < numToSkip)
            inStream.read();
	}

	/**
	 * Get a four byte big endian integer from the given
     * stream.
	 *
     * @param  inStream  Input stream from which to fetch
     *                   the integer.
     * @param  intBytes  Byte array into which to read the
     *                   bytes from <code>inStream</code>.
     * @return Integer value.
     * @throws Exception if an error occurs.
	 */
    protected int getBigEndianIntegerFromStream(InputStream inStream,
                                                byte[] intBytes) throws Exception {

        // Attempt to read the bytes all at once, but
        // read any individually that do not get read the
        // first time around (since the stream may not
        // have the bytes available to read the first
        // time around).
        int num = (int) inStream.read(intBytes);
        while (num < 4)
            intBytes[num++] = (byte) inStream.read();

        // Return the integer.
        return (intBytes[3] & 0xff) | (intBytes[2] & 0xff) << 8 |
            (intBytes[1] & 0xff) << 16 | intBytes[0] << 24;
	}

	/**
	 * Get a four byte little endian integer from the given
     * stream.
	 *
     * @param  inStream  Input stream from which to fetch
     *                   the integer.
     * @param  intBytes  Byte array into which to read the
     *                   bytes from <code>inStream</code>.
     * @return Integer value.
     * @throws Exception if an error occurs.
	 */
    protected int getLittleEndianIntegerFromStream(InputStream inStream,
                                                   byte[] intBytes) throws Exception {

        // Attempt to read the bytes all at once, but
        // read any individually that do not get read the
        // first time around (since the stream may not
        // have the bytes available to read the first
        // time around).
        int num = (int) inStream.read(intBytes);
        while (num < 4)
            intBytes[num++] = (byte) inStream.read();

        // Return the integer.
        return (intBytes[0] & 0xff) | (intBytes[1] & 0xff) << 8 |
            (intBytes[2] & 0xff) << 16 | intBytes[3] << 24;
	}

	/**
	 * Get an eight byte little endian double from the given
     * stream.
	 *
     * @param  inStream  Input stream from which to fetch
     *                   the double.
     * @param  intBytes  Byte array into which to read the
     *                   bytes from <code>inStream</code>.
     * @return Integer value.
     * @throws Exception if an error occurs.
	 */
    protected double getLittleEndianDoubleFromStream(InputStream inStream,
                                                     byte[] doubleBytes) throws Exception {

        // Attempt to read the bytes all at once, but
        // read any individually that do not get read the
        // first time around (since the stream may not
        // have the bytes available to read the first
        // time around).
        int num = (int) inStream.read(doubleBytes);
        while (num < 8)
            doubleBytes[num++] = (byte) inStream.read();

        // Return the double.
        return Double.longBitsToDouble((long) (doubleBytes[0] & 0xff) |
                                       (long) (doubleBytes[1] & 0xff) << 8 |
                                       (long) (doubleBytes[2] & 0xff) << 16 |
                                       (long) (doubleBytes[3] & 0xff) << 24 |
                                       (long) (doubleBytes[4] & 0xff) << 32 |
                                       (long) (doubleBytes[5] & 0xff) << 40 |
                                       (long) (doubleBytes[6] & 0xff) << 48 |
                                       (long) (doubleBytes[7] & 0xff) << 56);
	}

    /**
     * Get a latitude-longitude coordinate pair from the
     * input stream.
     *
     * @param  inStream    Input stream from which to fetch
     *                     the coordinate pair.
     * @param  doubleBytes Byte array into which to read
     *                     the bytes from <code>
     *                     inStream</code>.
     * @return Latitude-longitude coordinate pair.
     * @throws Exception if an error occurs.
     */
    public LatLong getLatLongFromStream(InputStream inStream,
                                        byte[] doubleBytes) throws Exception {
        double x = getLittleEndianDoubleFromStream(inStream, doubleBytes);
        double y = getLittleEndianDoubleFromStream(inStream, doubleBytes);
        return new LatLong((float) y, (float) x);
    }

    /**
     * Read the DBF file from the specified input stream.
     *
     * @param  inStream Input stream from which to read
     *                  the DBF file data.
     * @param  size     Size of the DBF file.
     * @return List of property values that was read in,
     *         each one for the shape(s) that have the
     *         same identifier as the index at which the
     *         values are found in the list.
     * @throws Exception if an error occurs.
     */
    protected ArrayList<Object[]> readDBF(InputStream inStream, long size) throws Exception {

        // Create the list to hold the property values.
        ArrayList<Object[]> propertyValues = new ArrayList<Object[]>();

        // Read in the DBF file.
        DBFReader dbfReader = new DBFReader(inStream);

        // Record the property names, and the index that
        // goes with each name.
        propertyNames = new String[dbfReader.getFieldCount()];
        propertyIndicesByNames = new HashMap<String, Integer>();
        for (int j = 0; j < propertyNames.length; j++) {
            propertyNames[j] = dbfReader.getField(j).getName();
            propertyIndicesByNames.put(propertyNames[j], j);
        }

        // Iterate through the values for each record,
        // remembering each.
        int maxValues = ShapefileDisplayProperties.getMaximumNumberTextValues();
        for (int count = 0; dbfReader.hasNextRecord(); count++)
            propertyValues.add(dbfReader.nextRecord());

        // Create default property mappings.
        propertyMappings = new ArrayList<ShapefileDisplayProperties.Property>();
        for (int j = 0; j < propertyNames.length; j++) {
            if (propertyValues.get(0)[j] instanceof Double) {
                boolean first = true;
                double[] values = { 0.0, 0.0 };
                for (int k = 0; k < propertyValues.size(); k++) {
                    double value = (Double) propertyValues.get(k)[j];
                    if (first) {
                        values[0] = values[1] = value;
                        first = false;
                    }
                    else if (value < values[0])
                        values[0] = value;
                    else if (value > values[1])
                        values[1] = value;
                }
                propertyMappings.add(new ShapefileDisplayProperties.
                                     FloatingProperty(propertyNames[j], values, COLOR));
            } else if (propertyValues.get(0)[j] instanceof Long) {
                boolean first = true;
                int[] values = { 0, 0 };
                for (int k = 0; k < propertyValues.size(); k++) {
                    int value = (int) ((Long) propertyValues.get(k)[j]).longValue();
                    if (first) {
                        values[0] = values[1] = value;
                        first = false;
                    }
                    else if (value < values[0])
                        values[0] = value;
                    else if (value > values[1])
                        values[1] = value;
                }
                propertyMappings.add(new ShapefileDisplayProperties.
                                     IntegerProperty(propertyNames[j], values, COLOR));
            } else {
                HashSet<String> values = new HashSet<String>();
                for (int k = 0; k < propertyValues.size(); k++)
                    values.add((String) propertyValues.get(k)[j]);
                if ((maxValues < 0) || (values.size() <= maxValues)) {
                    String[] valuesArray = new String[values.size()];
                    Iterator<String> iter = values.iterator();
                    for (int k = 0; iter.hasNext(); k++)
                        valuesArray[k] = iter.next();
                    Arrays.sort(valuesArray);
                    propertyMappings.add(new ShapefileDisplayProperties.
                                         TextProperty(propertyNames[j], valuesArray, COLOR));
                }
            }
        }

        // Return the compiled list of property values.
        return propertyValues;
    }

    /**
     * Paint either high- or lower-priority annotations.
     *
     * @param  highPriority   Flag indicating whether
     *                        the higher priority or lower
     *                        priority annotations should
     *                        be painted.
     * @param  graphics       Graphics object to be used.
     * @param  visualInfo     Visual information about
     *                        this map or product, if any.
     * @param  depictProps    Display properties to be
     *                        used, if any.
     * @param  observer       Image observer that is to be
     *                        notified if a paint cannot
     *                        be completed because an
     *                        image has not yet finished
     *                        loading.
     * @param  annotationGrid Annotation grid indicating
     *                        which cells are occupied
     *                        and which are not. This
     *                        grid is to be updated with
     *                        any cells into which this
     *                        method paints annotations
     *                        marked as used when this
     *                        method completes execution.
     * @return If painting could not be completed because
     *         a required image has not yet finished
     *         loading, the image that is still loading
     *         is returned. This implementation always
     *         returns <code>null</code>.
     */
    protected Image paintAnnotations(boolean highPriority, Graphics graphics,
                                     ProductVisualInfo visualInfo,
                                     DepictableDisplayProperties depictProps,
                                     ImageObserver observer,
                                     AnnotationGrid annotationGrid) {

        // If the cached labels have not been generated,
        // or if labels are not to be drawn, do nothing.
        if ((cachedHiPriLabelStrings == null) ||
            (((ShapefileDisplayProperties) depictProps).isShowingLabels() == false))
            return null;

        // Set the color and font to be used for the
        // labels, and set antialiasing if asked for.
        graphics.setColor(properties.transform(((ShapefileDisplayProperties) depictProps).
                                               getFontColor()));
        graphics.setFont(cachedFont);
        Object antiAliasingHint = null;
        if (System.getProperty("fxc.display.smoothing", "false").toLowerCase().equals("true")) {
            antiAliasingHint = ((Graphics2D) graphics).
                getRenderingHint(RenderingHints.KEY_ANTIALIASING);
            ((Graphics2D) graphics).setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                                                     RenderingHints.VALUE_ANTIALIAS_ON);
        }

        // Determine which cached set of annotations 
        // is to be drawn.
        ArrayList<String> cachedLabelStrings =
            (highPriority ? cachedHiPriLabelStrings : cachedOtherLabelStrings);
        ArrayList<Double> cachedLabelAngles =
            (highPriority ? cachedHiPriLabelAngles : cachedOtherLabelAngles);
        ArrayList<Point> cachedLabelPoints =
            (highPriority ? cachedHiPriLabelPoints : cachedOtherLabelPoints);
        ArrayList<HashSet<Point>> cachedLabelCells =
            (highPriority ? cachedHiPriLabelCells : cachedOtherLabelCells);
        ArrayList<Point> cachedLabelVisualEmphasisPoints =
            (highPriority ? cachedHiPriLabelVisualEmphasisPoints : null);

        // Iterate through the cached labels, drawing
        // each one in turn if the cells it would
        // occupy are not already occupied.
        for (int j = 0; j < cachedLabelStrings.size(); j++) {

            // Make sure that the cells that this label
            // would occupy are not already in use.
            Iterator<Point> cellIterator = cachedLabelCells.get(j).iterator();
            boolean alreadyOccupied = false;
            while (cellIterator.hasNext()) {
                Point cell = cellIterator.next();
                if (annotationGrid.isCellOccupied(cell.x, cell.y)) {
                    alreadyOccupied = true;
                    break;
                }
            }
            if (alreadyOccupied)
                continue;

            // Mark the cells that this label will
            // occupy as used.
            cellIterator = cachedLabelCells.get(j).iterator();
            while (cellIterator.hasNext()) {
                Point cell = cellIterator.next();
                annotationGrid.setCellOccupied(cell.x, cell.y);
            }

            // Draw the text using the specified
            // rotation and location.
            double adjustedAngle = cachedLabelAngles.get(j);
            Point startPoint = cachedLabelPoints.get(j);
            ((Graphics2D) graphics).rotate(adjustedAngle, startPoint.x,
                                           startPoint.y);
            graphics.drawString(cachedLabelStrings.get(j), startPoint.x,
                                startPoint.y);
            ((Graphics2D) graphics).rotate(-adjustedAngle, startPoint.x,
                                           startPoint.y);

            // Draw the point to be visually empha-
            // sized, if any, in the same color.
            if ((cachedLabelVisualEmphasisPoints != null) &&
                ((startPoint = cachedLabelVisualEmphasisPoints.get(j)) != null)) {
                Ellipse2D.Float ellipse = new Ellipse2D.Float(startPoint.x - 2,
                                                              startPoint.y - 2,
                                                              4, 4);
                ((Graphics2D) graphics).draw(ellipse);
                ((Graphics2D) graphics).fill(ellipse);
            }
        }

        // Unset the antialiasing in case it was set
        // above.
        if (antiAliasingHint != null)
            ((Graphics2D) graphics).setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                                                     antiAliasingHint);

        // Always return null, as no waiting for an
        // image needs to be done.
        return null;
    }

    /**
     * Flush any display-specific cached information
     * in response to a display parameter changing.
     * This implementation of this method does nothing.
     */
    protected void flushDisplayDependentData() {
        cachedDrawables = null;
        cachedPointsInCells = null;
        cachedPathsInCells = null;
        cachedCentroidsInCells = null;
        cachedPolygonsInCells = null;
        cachedPolygons = null;
        cachedHiPriLabelStrings = null;
        cachedHiPriLabelAngles = null;
        cachedHiPriLabelPoints = null;
        cachedHiPriLabelCells = null;
    }

    /**
     * Trim the geodata name to remove bogus characters.
     *
     * @param  name Original name returned by <code>
     *         GeoData.getName()</code>.
     * @return Name with bogus characters trimmed off.
     */
    protected String trimGeoDataName(String name) {
        for (int j = 0; j < name.length(); j++) {
            if (name.charAt(j) == (char) 0) {
                name = name.substring(0, j);
                break;
            }
        }
        return name;
    }

    /**
     * Get the annotation string describing the specified
     * shape, if any.
     *
     * @param  constants List of annotation constants,
     *                   into which are interleaved the
     *                   values from the fields in <code>
     *                   variables</code>; this array must
     *                   be at least one element in size,
     *                   and must be one element greater
     *                   size than <code>variables</code>.
     * @param  variables List of annotation variables.
     * @param  geoShape  Shape for which the text is being
     *                   generated.
     * @return Annotation string for this shape, or an
     *         empty string if none is available.
     */
    protected String getAnnotationTextForShape(String[] constants,
                                               String[] variables,
                                               GeoShape geoShape) {
        String text = "";
        for (int j = 0; j < constants.length - 1; j++) {
            int nameIndex = propertyIndicesByNames.get(variables[j]);
            Object value = geoShape.getProperties()[nameIndex];
            String fieldValue = null;
            if (value == null)
                fieldValue = "";
            else
                fieldValue = value.toString().trim();
            text = text + constants[j] + fieldValue;
        }
        text = text + constants[constants.length - 1];
        return text.trim();
    }

    /**
     * Get the centroid for the specified polygon.
     *
     * @param  points Points making up the polygon, with
     *                each ring in the polygon having the
     *                last point the same as the first
     *                point.
     * @return Centroid for the polygon.
     */
    protected Point getCentroidOfPolygon(ArrayList<Point> points) {

        // This calculation comes from the formulas at
        // http://local.wasp.uwa.edu.au/~pbourke/geometry/polyarea
        // It assumes that the polygon does not inter-
        // sect with itself, which it should not as per
        // the shapefile specs.`
        double area = 0.0;
        int sumX = 0, sumY = 0;
        for (int j = 0; j < points.size() - 1; j++) {
            Point point1 = points.get(j);
            Point point2 = points.get(j + 1);
            int multiplier = (point1.x * point2.y) - (point2.x * point1.y);
            sumX += (point1.x + point2.x) * multiplier;
            sumY += (point1.y + point2.y) * multiplier;
            area += multiplier;
        }
        area *= 3.0;
        return new Point((int) ((((double) sumX) / (area)) + 0.5),
                         (int) ((((double) sumY) / (area)) + 0.5));
    }

    /**
     * Get the basic bounding rectangle (centered on 0,0)
     * for the specified text.
     *
     * @param  text Text for which the bounding rectangle
     *              is needed.
     * @param  fm   Font metrics to be used for the
     *              bounding rectangle calculation.
     * @return Bounding rectangle (centered on 0,0) for
     *         the text.
     */
    protected Rectangle getBasicBoundsOfText(String text, FontMetrics fm) {
        Rectangle boundingRect = new Rectangle();
        boundingRect.width = fm.stringWidth(text);
        boundingRect.height = fm.getMaxAscent() + fm.getMaxDescent();
        boundingRect.x = boundingRect.width / -2;
        boundingRect.y = boundingRect.height / -2;
        return boundingRect;
    }

    /**
     * Determine whether or not the axes-aligned (i.e.
     * not rotated) rectangle specified may be drawn
     * without intruding into cells that are already
     * marked as occupied.
     *
     * @param  boundingRect   Bounding rectangle to be
     *                        checked.
     * @param  cellWidth      Width of each cell (except
     *                        possibly the rightmost one
     *                        in a line) in pixels.
     * @param  cellHeight     Height of each cell (except
     *                        possibly the bottommost one
     *                        in a column) in pixels.
     * @param  cellsUsed      Grid of flags indicating
     *                        whether or not each cell is
     *                        already occupied, indexed by
     *                        cell X and Y.
     * @param  centerCell     Cell that is the text center.
     * @param  cellPolygons   Optional argument that, if
     *                        provided, will cause a check
     *                        of all cells that the
     *                        rectangle occupies to ensure
     *                        that each such cell is marked
     *                        as for <code>polygon</code>.
     * @param  polygon        Required only if <code>
     *                        cellPolygons</code> is also
     *                        provided, in which case it is
     *                        used to verify each cell that
     *                        the rectangle falls within as
     *                        being marked as for this
     *                        polygon.
     * @return Set of cells that have been marked as used
     *         by this rectangle if the axes-aligned (i.e.
     *         not rotated) rectangle may be drawn without
     *         intruding into cells that are already marked
     *         as occupied, or <code>null</code> if this is
     *         not possible. If the former, then <code>
     *         cellsUsed</code> will have been altered to
     *         indicate that the rectangle is occupying the
     *         appropriate cells.
     */
    protected HashSet<Point> canAxesAlignedRectangleBeDrawn(Rectangle boundingRect,
                                                            int cellWidth, int cellHeight,
                                                            boolean[][] cellsUsed,
                                                            Point centerCell,
                                                            AnnotatedPolygon[][] cellPolygons,
                                                            AnnotatedPolygon polygon) {

        // If the center cell is already occupied, return
        // nothing.  Otherwise, record this cell as to be
        // used if this rectangle is found to be okay to
        // draw.  Also return false if polygon cell
        // relationships are being checked and this cell
        // is not marked as being for ths polygon.
        if ((centerCell.x >= 0) && (centerCell.x < cellsUsed.length) &&
            (centerCell.y >= 0) && (centerCell.y < cellsUsed[0].length) &&
            (cellsUsed[centerCell.x][centerCell.y] ||
             ((cellPolygons != null) &&
              (cellPolygons[centerCell.x][centerCell.y] != polygon))))
            return null;
        HashSet<Point> cellsToBeUsed = new HashSet<Point>();
        if ((centerCell.x >= 0) && (centerCell.x < cellsUsed.length) &&
            (centerCell.y >= 0) && (centerCell.y < cellsUsed[0].length))
            cellsToBeUsed.add(centerCell);

        // Determine the minimum and maximum cell bounda-
        // ries in both X and Y directions.
        int minCellX = boundingRect.x / cellWidth;
        if (minCellX < 0)
            minCellX = 0;
        int maxCellX = (boundingRect.x + boundingRect.width) / cellWidth;
        if (maxCellX >= cellsUsed.length)
            maxCellX = cellsUsed.length - 1;
        int minCellY = boundingRect.y / cellHeight;
        if (minCellY < 0)
            minCellY = 0;
        int maxCellY = (boundingRect.y + boundingRect.height) / cellHeight;
        if (maxCellY >= cellsUsed[0].length)
            maxCellY = cellsUsed[0].length - 1;

        // Iterate through the cells within those
        // boundaries, including the boundary cells
        // themselves, checking to ensure that each is
        // not already occupied, and build up a set of
        // cells that this rectangle would occupy.
        // Return nothing if a cell is encountered that
        // is already occupied, or if polygon cell
        // relationships are being checked and this
        // cell is not marked as being for ths polygon.
        for (int y = minCellY; y <= maxCellY; y++) {
            for (int x = minCellX; x <= maxCellX; x++) {
                if (cellsUsed[x][y] ||
                    ((cellPolygons != null) &&
                     (cellPolygons[x][y] != polygon)))
                    return null;
                cellsToBeUsed.add(new Point(x, y));
            }
        }

        // Since none of the cells are already occupied,
        // mark them as occupied now and return the set
        // of these cells to indicate that the rectangle
        // fits.
        Iterator<Point> cellIterator = cellsToBeUsed.iterator();
        while (cellIterator.hasNext()) {
            Point cell = cellIterator.next();
            cellsUsed[cell.x][cell.y] = true;
        }
        return cellsToBeUsed;
    }

    /**
     * Determine whether or not the specified arbitrarily
     * rotated rectangle may be drawn without intruding
     * into cells that are already marked as occupied.
     *
     * @param  boundsIterator Iterator that provides the
     *                        bounding points of the
     *                        arbitrarily rotated rectangle.
     *                        The iterator must return
     *                        exactly four segments, each
     *                        assumed to be of type <code>
     *                        SEG_MOVETO</code> or <code>
     *                        SEG_LINETO</code>.
     * @param  cellWidth      Width of each cell (except
     *                        possibly the rightmost one
     *                        in a line) in pixels.
     * @param  cellHeight     Height of each cell (except
     *                        possibly the bottommost one
     *                        in a column) in pixels.
     * @param  cellsUsed      Grid of flags indicating
     *                        whether or not each cell is
     *                        already occupied, indexed by
     *                        cell X and Y.
     * @param  centerCell     Cell that is the text
     *                        center.
     * @return Set of cells that have been marked as used
     *         by this rectangle if the arbitrarily rotated
     *         rectangle may be drawn without intruding
     *         into cells that are already marked as
     *         occupied, or <code>null</code> if this is
     *         not possible. If the former, then <code>
     *         cellsUsed</code> will have been altered to
     *         indicate that the rectangle is occupying the
     *         appropriate cells.
     * @see java.awt.geom.PathIterator
     */
    protected HashSet<Point> canArbitraryRectangleBeDrawn(PathIterator boundsIterator,
                                                          int cellWidth, int cellHeight,
                                                          boolean[][] cellsUsed,
                                                          Point centerCell) {

        // If the center cell is already occupied, return
        // nothing.  Otherwise, record this cell as to be
        // used if this rectangle is found to be okay to
        // draw.
        if ((centerCell.x >= 0) && (centerCell.x < cellsUsed.length) &&
            (centerCell.y >= 0) && (centerCell.y < cellsUsed[0].length) &&
            cellsUsed[centerCell.x][centerCell.y])
            return null;
        HashSet<Point> cellsToBeUsed = new HashSet<Point>();
        if ((centerCell.x >= 0) && (centerCell.x < cellsUsed.length) &&
            (centerCell.y >= 0) && (centerCell.y < cellsUsed[0].length))
            cellsToBeUsed.add(centerCell);

        // Get the points making up the rotated rectangle,
        // and check each one to see which cell it is in;
        // if that cell is already occupied, return
        // nothing.  Otherwise, remember the cell for use
        // later.
        float[] coords = new float[6];
        Point[] boundingPoints = new Point[4];
        Point[] boundingCells = new Point[4];
        int minCellY = 1000000, maxCellY = -1000000;
        for (int m = 0; m < 4; m++) {

            // Determine the point and the cell that this
            // point lies within.
            boundsIterator.currentSegment(coords);
            boundingPoints[m] = new Point((int) (coords[0] + 0.5f),
                                          (int) (coords[1] + 0.5f));
            int thisX = boundingPoints[m].x / cellWidth;
            int thisY = boundingPoints[m].y / cellHeight;

            // See if this cell is above or below the pre-
            // viously-recorded minimum and maximum cellu-
            // lar Y values, and if so, use this value as
            // the new minimum and/or maximum Y cellular
            // value.
            if (minCellY > thisY)
                minCellY = thisY;
            if (maxCellY < thisY)
                maxCellY = thisY;

            // If the cell is not outside the bounds of
            // the display, ensure it is not already
            // occupied.
            if ((thisX >= 0) && (thisX < cellsUsed.length) &&
                (thisY >= 0) && (thisY < cellsUsed[0].length) &&
                cellsUsed[thisX][thisY])
                return null;

            // Remember the cell for more checking below,
            // and if it lies within the bounds of the
            // display, add it to the list of cells to be
            // marked as used by this rectangle.
            boundingCells[m] = new Point(thisX, thisY);
            if ((thisX >= 0) && (thisX < cellsUsed.length) &&
                (thisY >= 0) && (thisY < cellsUsed[0].length))
                cellsToBeUsed.add(boundingCells[m]);
            boundsIterator.next();
        }

        // Create an array for minimum X cellular values,
        // and another for maximum X cellular values, to
        // determine, for each cellular line that this
        // rectangle may occupy, the minimum and maximum
        // X values found on that line.  The value of
        // minCellY will be used as an offset, so that
        // minCellX[2] will be for line minCellY + 2.
        int[] minCellX = new int[maxCellY + 1 - minCellY];
        int[] maxCellX = new int[maxCellY + 1 - minCellY];
        for (int m = 0; m < minCellX.length; m++) {
            minCellX[m] = 1000000;
            maxCellX[m] = -1000000;
        }

        // Mark the four points of the rotated bounding
        // rectangle in the minimum/maximum cellular X
        // value arrays.
        for (int m = 0; m < 4; m++) {
            if (minCellX[boundingCells[m].y - minCellY] >
                boundingCells[m].x)
                minCellX[boundingCells[m].y - minCellY] =
                    boundingCells[m].x;
            if (maxCellX[boundingCells[m].y - minCellY] <
                boundingCells[m].x)
                maxCellX[boundingCells[m].y - minCellY] =
                    boundingCells[m].x;
        }

        // For each segment, determine what cells it
        // passes through between the endpoint cells, and
        // whether these intermediate cells are occupied.
        // If so, return nothing.
        for (int m = 0; m < 4; m++) {

            // If the segment is all within one cell, it
            // has already been tested, so don't test it
            // again.  If the cells are all in the same
            // line, just test all cells in a horizontal
            // line between them; likewise, if they are
            // in the same column, test the intermediate
            // cells in that column.  Otherwise, do a
            // more detailed check to determine what
            // cells the segment passes through between
            // its endpoints.
            Point cell1 = boundingCells[m];
            Point cell2 = boundingCells[m == 3 ? 0 : m + 1];
            if (cell1.equals(cell2))
                continue;
            else if (cell1.y == cell2.y) {
                int offset = (cell1.x < cell2.x ? 1 : -1);
                for (int thisX = cell1.x + offset; thisX != cell2.x;
                     thisX += offset) {

                    // If the line is outside the bounds
                    // of the display, no more checking
                    // needs to be done; it is safe.
                    if ((cell1.y < 0) || (cell1.y >= cellsUsed[0].length))
                        break;

                    // If the column is outside the
                    // bounds of the display, it does
                    // not need to be checked; it is
                    // safe.
                    if ((thisX < 0) || (thisX >= cellsUsed.length))
                        continue;

                    // Ensure the cell is not already
                    // occupied.
                    if (cellsUsed[thisX][cell1.y])
                        return null;

                    // Add it to the list of cells to
                    // be marked as used by this rec-
                    // tangle
                    cellsToBeUsed.add(new Point(thisX, cell1.y));
                }
            } else if (cell1.x == cell2.x) {
                int offset = (cell1.y < cell2.y ? 1 : -1);
                for (int thisY = cell1.y + offset; thisY != cell2.y;
                     thisY += offset) {

                    // If the column is outside the
                    // bounds of the display, no more
                    // checking needs to be done; it
                    // is safe.
                    if ((cell1.x < 0) || (cell1.x >= cellsUsed.length))
                        break;

                    // If the column lies outside the
                    // minimum and maximum cellular X
                    // values found thus far for this
                    // line, record it as the minimum
                    // and/or maximum value.
                    if (minCellX[thisY - minCellY] > cell1.x)
                        minCellX[thisY - minCellY] = cell1.x;
                    if (maxCellX[thisY - minCellY] < cell1.x)
                        maxCellX[thisY - minCellY] = cell1.x;

                    // If the line is outside the
                    // bounds of the display, it does
                    // not need to be checked; it is
                    // safe.
                    if ((thisY < 0) || (thisY >= cellsUsed[0].length))
                        continue;

                    // Ensure the cell is not already
                    // occupied.
                    if (cellsUsed[cell1.x][thisY])
                        return null;

                    // Add it to the list of cells to
                    // be marked as used by this rec-
                    // tangle.
                    cellsToBeUsed.add(new Point(cell1.x, thisY));
                }
            } else {

                // Calculate the slope and offset for
                // the equation of the line segment.
                Point point1 = boundingPoints[m];
                Point point2 = boundingPoints[m == 3 ? 0 : m + 1];
                double slope = ((double) (point2.y - point1.y)) /
                    (double) (point2.x - point1.x);
                double offset = ((double) point1.y) -
                    (slope * (double) point1.x);

                // Iterate through the cells that are
                // between the first and last one, mo-
                // ving along the line segment between
                // the start and end cells, ensuring
                // that each such cell is not occupied.
                int cellOffX = (cell1.x < cell2.x ? 1 : -1);
                int cellOffY = (cell1.y < cell2.y ? 1 : -1);
                Point cell = new Point(cell1);
                while (true) {

                    // Find the pixel line and column
                    // that define the horizontal and
                    // vertical boundaries of the cur-
                    // rent cell in the direction in
                    // which intersection with the line
                    // segment must be tested.
                    int boundX = (cell.x + (cellOffX == 1 ? 1 : 0)) *
                        cellWidth;
                    int boundY = (cell.y + (cellOffY == 1 ? 1 : 0)) *
                        cellHeight;

                    // Determine the intersection point
                    // for the line segment with each
                    // of these boundaries.
                    int y2 = (int) ((slope * (double) boundX) + offset + 0.5);
                    int x2 = (int) (((((double) boundY) - offset) / slope) + 0.5);

                    // If the line segment intersects
                    // the boundaries at the same
                    // point, then the next cell is
                    // located diagonally from this one.
                    // Otherwise, if the boundary passed
                    // through first is the vertical
                    // one, then the next cell is loca-
                    // ted to one side horizontally.  If
                    // neither of the above are the
                    // case, then the next cell is loca-
                    // ted above or below this one.
                    if ((boundX == x2) && (boundY == y2)) {
                        cell.x += cellOffX;
                        cell.y += cellOffY;
                    } else if (((cellOffX == 1) && (boundX < x2)) ||
                               ((cellOffY == 1) && (y2 < boundY)) ||
                               ((cellOffX == -1) && (boundX > x2)) ||
                               ((cellOffY == -1) && (y2 > boundY)))
                        cell.x += cellOffX;
                    else
                        cell.y += cellOffY;

                    // Make sure that the iteration is
                    // not passing into or beyond the
                    // end cell.
                    if (cell.equals(cell2) ||
                        (cell.x == cell2.x + cellOffX) ||
                        (cell.y == cell2.y + cellOffY))
                        break;

                    // If the column lies outside the
                    // minimum or maximum cellular X
                    // values found thus far for this
                    // line, record it as the minimum
                    // and/or maximum value.
                    if (minCellX[cell.y - minCellY] > cell.x)
                        minCellX[cell.y - minCellY] = cell.x;
                    if (maxCellX[cell.y - minCellY] < cell.x)
                        maxCellX[cell.y - minCellY] = cell.x;

                    // If the cell is not within the
                    // boundaries of the display, it
                    // does not need to be checked;
                    // it is safe.
                    if ((cell.x < 0) || (cell.x >= cellsUsed.length) ||
                        (cell.y < 0) || (cell.y >= cellsUsed[0].length))
                        continue;
                                
                    // Ensure the cell is not already
                    // occupied.
                    if (cellsUsed[cell.x][cell.y])
                        return null;

                    // Add it to the list of cells to
                    // be marked as used by this rec-
                    // tangle.
                    cellsToBeUsed.add(new Point(cell));
                }
            }
        }

        // Make sure that all the cells that might lie
        // in between the segments making up the boun-
        // ding rectangle are also free; if they are
        // not, return nothing.
        for (int m = 1; m < minCellX.length - 1; m++) {
            for (int n = minCellX[m] + 1; n < maxCellX[m]; n++) {

                // If this cell is outside the bounda-
                // ries of the display, it does not
                // need to be checked; it is safe.
                if ((n < 0) || (n >= cellsUsed.length) ||
                    (m + minCellY < 0) || (m + minCellY >=
                                           cellsUsed[0].length))
                    continue;

                // Ensure the cell is not already
                // occupied.
                if (cellsUsed[n][m + minCellY])
                    return null;

                // Add it to the list of cells to be
                // marked as used by this rectangle.
                cellsToBeUsed.add(new Point(n, m + minCellY));
            }
        }

        // Since none of the cells are already occupied,
        // mark them as occupied now and return the set
        // of cells occupied by this rectangle in order
        // to indicate that the latter fits.
        Iterator<Point> cellIterator = cellsToBeUsed.iterator();
        while (cellIterator.hasNext()) {
            Point cell = cellIterator.next();
            if ((cell.x >= 0) && (cell.x < cellsUsed.length) &&
                (cell.y >= 0) && (cell.y < cellsUsed[0].length))
                cellsUsed[cell.x][cell.y] = true;
        }
        return cellsToBeUsed;
    }

    /**
     * Determine whether the specified path intersects the
     * specified rectangle (or is contained fully within
     * the latter).
     *
     * @param  path      List of points making up the path.
     * @param  rectangle Rectangle.
     * @return True if the shapes intersect, otherwise
     *         false.
     */
    protected boolean pathIntersectsRectangle(ArrayList<Point> path,
                                              Rectangle rectangle) {
        for (int j = 0; j < path.size() - 1; j++) {
            Point point1 = path.get(j);
            Point point2 = path.get(j + 1);
            if (rectangle.intersectsLine(point1.x, point1.y, point2.x, point2.y))
                return true;
        }
        return false;

        /*
        // First, determine whether any of the path's
        // points lie within the rectangle as a shortcut;
        // if one does, return true.
        for (int j = 0; j < path.size(); j++) {
            Point point = path.get(j);
            if (rectangle.contains(point.x, point.y))
                return true;
        }

        // Since no points were found that lie within the
        // rectangle, the segments must be checked to see
        // if they pass through or along the edge of the
        // rectangle.  First, make a list of the four
        // corner points of the rectangle.
        Point[] rectPoints = new Point[4];
        for (int j = 0; j < 4; j++)
            rectPoints[j] =
                new Point((j < 2 ? rectangle.x : rectangle.x + rectangle.width),
                          (j % 2 == 0 ? rectangle.y : rectangle.y + rectangle.height));

        // Next, iterate through the line segments,
        // checking each in turn to see if it intersects
        // with the rectangle.
        for (int j = 0; j < path.size() - 1; j++) {

            // Given that x,y is a corner point for the
            // rectangle, x1,y1 is one end of the seg-
            // ment, and x2,y2 is the other, the formula
            // (y2-y1)x + (x1-x2)y + (x2*y1) -(x1*y2)
            // is applied to each of the four corner
            // points, and the results compared.  If it
            // ever yields 0, a segment endpoint lies
            // along the edge of the rectangle.  Other-
            // wise, if all four iterations of the
            // application of this formula yield results
            // with the same sign (all positive or all
            // negative), the segments do not intersect.
            Point point1 = path.get(j);
            Point point2 = path.get(j + 1);
            int xFactor = point2.y - point1.y;
            int yFactor = point1.x - point2.x;
            int offset = (point2.x * point1.y) - (point1.x * point2.y);
            int lastResult = 0;
            for (int k = 0; k < rectPoints.length; k++) {
                int thisResult =
                    (xFactor * rectPoints[k].x) + (yFactor * rectPoints[k].y) + offset;
                if (thisResult == 0)
                    return true;
                else if (k == 0)
                    lastResult = thisResult;
                else if (thisResult * lastResult < 0)
                    return true;
            }
        }

        // If no intersection was found, return false.
        return false;
        */
    }

    /**
     * Find the point on the path that is closest to the
     * specified point.
     *
     * @param  point   A single point that may or may not
     *                 lie on the path.
     * @param  path    List of points defining the path.
     * @param  section Length of the section of the path to
     *                 be used to calculate the angle. The
     *                 section will have half its length on
     *                 each side of the point if possible.
     * @param  nearest Point in which the coordinates of the
     *                 point along the path is stored upon
     *                 execution of the method; nothing
     *                 meaningful is stored if -1.0 is
     *                 returned).
     * @return The angle in degrees of the point stored in
     *         <code>nearest</code>, or -1.0 if no suitable
     *         point could be found.
     */
    protected double findNearestPointAndAngleOnPath(Point point, ArrayList<Point> path,
                                                    int section, Point nearest) {

        // Find the segment to which the point is closest.
        double closestDistance = 1000000.0;
        int closestSegment = -1;
        for (int j = 1; j < path.size(); j++) {

            // If the distance between the point and this
            // path segment is the closest so far, record it
            // and the segment's index.
            Point point1 = path.get(j - 1);
            Point point2 = path.get(j);
            double dist = Line2D.ptSegDist(point1.x, point1.y, point2.x, point2.y,
                                           point.x, point.y);
            if (dist < closestDistance) {
                closestDistance = dist;
                closestSegment = j - 1;
            }
        }

        // Find the point along the closest segment that is
        // closest to the specified point.
        if (closestSegment != -1) {
            Point point1 = path.get(closestSegment);
            Point point2 = path.get(closestSegment + 1);

            // If the segment is vertical or horizontal,
            // the point is simple to calcluate; otherwise,
            // the point's coordinates require the calcu-
            // lation first of the segment's slope and off-
            // set (m and b from y = mx + b), then the slope
            // and offset of the line running through the
            // original point perpendicularly to the seg-
            // ment, and finally the finding of an inter-
            // section between these two lines.
            if (point2.x == point1.x) {
                nearest.x = point1.x;
                nearest.y = point.y;

                // Make sure that the y coordinate is within
                // the bounds of the segment's y coordinates.
                if (nearest.y < (point1.y < point2.y ? point1.y : point2.y))
                    nearest.y = (point1.y < point2.y ? point1.y : point2.y);
                if (nearest.y > (point1.y > point2.y ? point1.y : point2.y))
                    nearest.y = (point1.y > point2.y ? point1.y : point2.y);
            } else if (point2.y == point1.y) {
                nearest.x = point.x;
                nearest.y = point1.y;

                // Make sure that the x coordinate is within
                // the bounds of the segment's x coordinates.
                if (nearest.x < (point1.x < point2.x ? point1.x : point2.x))
                    nearest.x = (point1.x < point2.x ? point1.x : point2.x);
                if (nearest.x > (point1.x > point2.x ? point1.x : point2.x))
                    nearest.x = (point1.x > point2.x ? point1.x : point2.x);
            } else {

                // Get the slopes and offsets for the line
                // equations describing the segment's line
                // and the line running perpendicularly from
                // the point to the segment.
                double segmentM = ((double) (point2.y - point1.y)) /
                    (double) (point2.x - point1.x);
                double segmentB = ((double) point1.y) - (segmentM *
                                                         (double) point1.x);
                double perpM = -1.0 / segmentM;
                double perpB = ((double) point.y) - (perpM *
                                                     (double) point.x);

                // Calculate the point of intersection be-
                // tween the two line equations.
                double nearestX = (perpB - segmentB) / (segmentM - perpM);
                nearest.x = (int) Math.round(nearestX);
                nearest.y = (int) Math.round((perpM * nearestX) + perpB);

                // Make sure that the point calculated lies
                // on the segment; if not, adjust it so that
                // it is whichever segment endpoint is clo-
                // sest to it.
                if (nearest.x < (point1.x < point2.x ? point1.x : point2.x)) {
                    nearest.x = (point1.x < point2.x ? point1.x : point2.x);
                    nearest.y = (point1.x < point2.x ? point1.y : point2.y);
                } else if (nearest.y < (point1.y < point2.y ? point1.y : point2.y)) {
                    nearest.y = (point1.y < point2.y ? point1.y : point2.y);
                    nearest.x = (point1.y < point2.y ? point1.x : point2.x);
                } else if (nearest.x > (point1.x > point2.x ? point1.x : point2.x)) {
                    nearest.x = (point1.x > point2.x ? point1.x : point2.x);
                    nearest.y = (point1.x > point2.x ? point1.y : point2.y);
                } else if (nearest.y > (point1.y > point2.y ? point1.y : point2.y)) {
                    nearest.y = (point1.y > point2.y ? point1.y : point2.y);
                    nearest.x = (point1.y > point2.y ? point1.x : point2.x);
                }
            }
        }

        // Get the distance that this point is along the
        // path; then get the points that are each half of
        // section length away from the point to either
        // side along the path.  These points will be used
        // to calculate the angle between them.  If the
        // point needs to be moved, so be it.
        double startDistance, endDistance, distance = 0.0;
        double halfLength = ((double) section) / 2.0;
        double[] distances = new double[path.size()];
        for (int j = 1; j < path.size(); j++) {
            Point point1 = path.get(j);
            Point point2 = path.get(j - 1);
            distances[j] = (j == 1 ? 0.0 : distances[j - 1]) +
                Math.sqrt(((point1.x - point2.x) *
                           (point1.x - point2.x)) +
                          ((point1.y - point2.y) *
                           (point1.y - point2.y)));
            if (j == closestSegment + 1)
                distance = (j == 1 ? 0.0 : distances[j - 1]) +
                    Math.sqrt(((nearest.x - point2.x) *
                               (nearest.x - point2.x)) +
                              ((nearest.y - point2.y) *
                               (nearest.y - point2.y)));
        }
        boolean movePoint = false;
        if (distances[distances.length - 1] <= (double) section)
            return -1.0;
        else if (distance - halfLength <= 0.0) {
            startDistance = 0.0;
            endDistance = (double) section;
            distance = (endDistance / 2.0);
            movePoint = true;
        } else if (distance + halfLength >= distances[distances.length - 1]) {
            endDistance = distances[distances.length - 1];
            startDistance = endDistance - (double) section;
            distance = endDistance - (((double) section) / 2.0);
            movePoint = true;
        } else {
            startDistance = distance - halfLength;
            endDistance = distance + halfLength;
        }
        if (movePoint) {
            Point newPoint = getPointAtDistanceAlongPath(distance, path, distances);
            nearest.x = newPoint.x;
            nearest.y = newPoint.y;
        }
        Point point1 = getPointAtDistanceAlongPath(startDistance, path, distances);
        Point point2 = getPointAtDistanceAlongPath(endDistance, path, distances);

        // Calculate the angle between the two points.
        if (point2.x == point1.x) {
            if (point2.y == point1.y)
                return 0.0;
            else
                return 270;
        } else {
            double slope = ((double) (point2.y - point1.y)) /
                (double) (point2.x - point1.x);
            double angle = Math.toDegrees(Math.atan(slope));
            if (angle < 0.0)
                angle += 360.0;
            if ((angle >= 90.0) && (angle < 270))
                angle += 180.0;
            return angle % 360.0;
        }
    }

    /**
     * Get the point that is the specified distance along the
     * length of the glyph.
     *
     * @param  distance      Distance along the length of the
     *                       path at which the point lies.
     * @param  path          List of points defining the path.
     * @param  pathDistances Array of distances each point in
     *                       the corresponding index of <code>
     *                       path</code> is away from the
     *                       first point in <code>path</code>.
     * @return Point that lies at the specified distance along
     *         the length of the path.
     */
    protected Point getPointAtDistanceAlongPath(double distance, ArrayList<Point> path,
                                                double[] pathDistances) {

        // See if the distance given is at the beginning or
        // the end of the line; if so, just return the appro-
        // priate point.
        if (distance == 0.0)
            return path.get(0);
        else if (distance == pathDistances[pathDistances.length - 1])
            return path.get(path.size() - 1);

        // Do a binary search of the distances in order to
        // find which node is at the distance specified, or if
        // no such node exists, between which two nodes the
        // point that is the specified distance along the
        // length of the path is found.
        boolean betweenNodes = false;
        int lower = 0, upper = pathDistances.length, index = pathDistances.length / 2;
        while (distance != pathDistances[index]) {
            if (distance > pathDistances[index]) {
                if (distance < pathDistances[index + 1]) {
                    betweenNodes = true;
                    break;
                } else
                    lower = index;
            } else
                upper = index;
            index = ((upper - lower) / 2) + lower;
        }

        // If the point is between the indexed node and the
        // next one, find it exactly; otherwise, return the
        // node itself.
        if (betweenNodes) {
            double fraction = (distance - pathDistances[index]) /
                (pathDistances[index + 1] - pathDistances[index]);
            return new Point(((int) (fraction * (path.get(index + 1).x - path.get(index).x))) +
                             path.get(index).x,
                             ((int) (fraction * (path.get(index + 1).y - path.get(index).y))) +
                             path.get(index).y);
        } else
            return path.get(index);
    }
}
