package fsl.depict.shapefile;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.JViewport;
import javax.swing.ListSelectionModel;
import javax.swing.SwingConstants;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.AbstractTableModel;
import fsl.depict.shapefile.ShapefileDisplayProperties;
import fsl.gui.ChoiceContent;
import fsl.gui.ColorTableCellRenderer;
import fsl.gui.DialogContent;
import fsl.gui.ModalDialog;
import fsl.gui.NumberRangePanel;
import fsl.gui.OptionPane;
import fsl.gui.PanelTitledPanel;
import fsl.gui.ParameterComponent;
import fsl.gui.ParameterUser;
import fsl.gui.SearchDocument;
import fsl.gui.ToggleDetailPanel;
import fsl.gui.UserCommunicator;
import fsl.util.DisplayPalette;
import fsl.util.Logger;

/**
 * The content pane for the shapefile display properties
 * dialog, to be used to change the display property
 * mappings of a shapefile depictable.
 *
 * @author Christopher Golden
 * @see fsl.depict.shapefile.ShapefileDepictable
 */
public class ShapefileDisplayPropertiesContent extends DialogContent {


    // Private Classes

    /**
     * The mapping table model class, a model that tracks
     * the value-to-color mappings. It implements the
     * search document's inner model interface in order
     * to allow searches to be performed for particular
     * property values when the property chosen is a
     * text property.
     */
    private class MappingTableModel extends AbstractTableModel
        implements SearchDocument.SearchableItemListModel {

        private JTable table = null;
        private ArrayList<String> values = new ArrayList<String>();
        private ArrayList<Boolean> showFlags = new ArrayList<Boolean>();
        private ArrayList<Color> colors = new ArrayList<Color>();

        public void setTable(JTable table) {
            this.table = table;
        }
        public int getRowCount() {
            return values.size();
        }
        public int getColumnCount() {
            return 3;
        }
        public String getColumnName(int column) {
            return (column == 0 ? "Value" : (column == 1 ? "Visible" : "Color"));
        }
        public Object getValueAt(int row, int column) {
            return (column == 0 ? values.get(row) :
                    (column == 1 ? showFlags.get(row) : colors.get(row)));
        }
        public Class getColumnClass(int column) {
            return (column == 0 ? String.class : (column == 1 ? Boolean.class :
                                                  Color.class));
        }
        public boolean isCellEditable(int row, int column) {
            return (column == 1);
        }
        public void setValueAt(Object value, int row, int column) {
            if (column == 0)
                values.set(row, (String) value);
            else if (column == 1) {
                showFlags.set(row, (Boolean) value);
                propertyVisibilityChanged(row, (Boolean) value);
            } else
                colors.set(row, (Color) value);
            fireTableCellUpdated(row, column);
        }
        public void clear() {
            values.clear();
            showFlags.clear();
            colors.clear();
            fireTableDataChanged();
        }
        public void populate(ArrayList<String> newValues,
                             ArrayList<Boolean> newShowFlags,
                             ArrayList<Color> newColors) {
            values = newValues;
            showFlags = newShowFlags;
            colors = newColors;
            fireTableDataChanged();
        }
        public int getSize() {
            return getRowCount();
        }
        public Object getElementAt(int index) {
            return getValueAt(index, 0);
        }
        public int getIndexOf(Object item) {
            String itemString = item.toString();
            int pos = 0, bottom = 0, top = getRowCount();
            while (bottom <= top) {
                pos = (bottom + top) / 2;
                int compare = itemString.compareTo(getValueAt(pos, 0).toString());
                if (compare == 0)
                    return pos;
                else if (compare > 0)
                    bottom = pos + 1;
                else
                    top = pos - 1;
            }
            return -1;
        }
        public int getSelectedIndex() {
            int[] rows = table.getSelectedRows();
            return ((rows != null) && (rows.length > 0) ? rows[0] : -1);
        }
        public Object getSelectedItem() {
            int index = getSelectedIndex();
            return (index == -1 ? null : getValueAt(index, 0));
        }
        public void setSelectedIndex(int index) {
            table.getSelectionModel().clearSelection();
            table.getSelectionModel().addSelectionInterval(index, index);
            JViewport viewport = (JViewport) table.getParent();
            Rectangle rect = table.getCellRect(index, 0, false);
            Rectangle viewRect = viewport.getViewRect();
            rect.setLocation(rect.x - viewRect.x, rect.y - viewRect.y);
            viewport.scrollRectToVisible(rect);
        }
        public void setSelectedItem(Object item) {
            int index = getIndexOf(item);
            if (index != -1)
                setSelectedIndex(index);
        }
    }

    /**
     * The searchable table class, a table that can be
     * searched using a <code>SearchDocument</code>.
     */
    private class SearchableTable extends JTable implements SearchDocument.SearchableComponent {

        public SearchableTable(MappingTableModel model) {
            super(model);
        }

        public void searchFailed() {
            getToolkit().beep();
        }
        public void searchKeyPressed() {
        }
    }

    /**
     * The change range dialog content pane, allowing an
     * arbitrary range to be created by the user.
     */
    private class ChangeRangeContent extends DialogContent implements ParameterUser {

        private class ColorSwatchPanel extends JPanel implements ParameterComponent {
            private JPanel colorSwatch;
            private JButton changeButton;

            public ColorSwatchPanel(Color color) {
                GridBagLayout gridLayout = new GridBagLayout();
                GridBagConstraints gc = new GridBagConstraints();
                setLayout(gridLayout);
                changeButton = new JButton("Change...");
                changeButton.addActionListener(new ActionListener() {
                        public void actionPerformed(ActionEvent e) {
                            Color color = colorSwatch.getBackground();
                            color = userCommunicator.
                                getUserColor(getDialogParent(ChangeRangeContent.this),
                                             color);
                            if (color != null)
                                colorSwatch.setBackground(color);
                        }
                    });
                colorSwatch = new JPanel();
                colorSwatch.
                    setPreferredSize(new Dimension(changeButton.getPreferredSize().height,
                                                   changeButton.getPreferredSize().height));
                colorSwatch.setBackground(color);
                colorSwatch.setBorder(BorderFactory.createLoweredBevelBorder());
                gc.gridx = 1;
                gc.gridy = 1;
                gridLayout.setConstraints(colorSwatch, gc);
                add(colorSwatch);
                JPanel separator = new JPanel();
                separator.setPreferredSize(new Dimension(10, 0));
                gc.gridx = 2;
                gridLayout.setConstraints(separator, gc);
                add(separator);
                gc.gridx = 3;
                gc.weightx = 1.0;
                gc.anchor = GridBagConstraints.WEST;
                gc.fill = GridBagConstraints.NONE;
                gridLayout.setConstraints(changeButton, gc);
                add(changeButton);
            }
            public Color getColor() {
                return colorSwatch.getBackground();
            }
            public void addUser(ParameterUser user) {
            }
            public boolean hasValidParameters() {
                return true;
            }
            public void enableOrDisable(boolean enable) {
                changeButton.setEnabled(enable);
            }
            public int getDescriptiveLabelWidth() {
                return 0;
            }
            public void setDescriptiveLabelWidth(int width) {
            }
        }

        private ToggleDetailPanel visibilityPanel;
        private NumberRangePanel numberRange;
        private ColorSwatchPanel colorSwatch;
        private boolean intValues;

        public ChangeRangeContent(int minimum, int maximum, int startingLower,
                                  int startingUpper, Boolean visible, Color startingColor) {
            numberRange = new NumberRangePanel(this, null, "Range:", " to ",
                                               minimum, maximum, 1, true,
                                               startingLower, startingUpper);
            intValues = true;
            initialize(startingColor, visible);
        }
        public ChangeRangeContent(double minimum, double maximum, double startingLower,
                                  double startingUpper, Boolean visible, Color startingColor) {
            double stepSize = (maximum - minimum > 100.0 ? 1.0 :
                               (maximum - minimum > 10.0 ? 0.1 :
                                (maximum - minimum > 1.0 ? 0.01 :
                                 (maximum - minimum > 0.1 ? 0.001 :
                                  (maximum - minimum > 0.01 ? 0.0001 :
                                   (maximum - minimum > 0.001 ? 0.00001 :
                                    (maximum - minimum > 0.0001 ? 0.000001 :
                                     (maximum - minimum > 0.00001 ? 0.0000001 :
                                      (maximum - minimum > 0.000001 ? 0.00000001 :
                                       (maximum - minimum > 0.0000001 ? 0.000000001 :
                                        (maximum - minimum > 0.00000001 ? 0.0000000001 :
                                         0.00000000001)))))))))));
            numberRange = new NumberRangePanel(this, null, "Range:", " \u2264 x < ",
                                               minimum, maximum, stepSize,
                                               false, startingLower, startingUpper,
                                               "0.################E0##");
            intValues = false;
            initialize(startingColor, visible);
        }

        public boolean validateInput(ModalDialog dialog) {
            if (numberRange.hasValidParameters())
                return true;
            OptionPane.showMessageDialog(dialog, "The range's lower bound must be " +
                                         (intValues ? "less than or equal to " :
                                          "less than ") + "the upper bound.");
            return false;
        }
        public void parameterValidityChanged(ParameterComponent component,
                                             boolean valid) {
        }
        public void parameterComponentChanged() {
        }
        public int getIntegerLower() {
            return numberRange.getIntegerLower();
        }
        public int getIntegerUpper() {
            return numberRange.getIntegerUpper();
        }
        public double getDoubleLower() {
            return numberRange.getDoubleLower();
        }
        public double getDoubleUpper() {
            return numberRange.getDoubleUpper();
        }
        public Boolean getVisible() {
            return visibilityPanel.getState();
        }
        public Color getColor() {
            return colorSwatch.getColor();
        }

        private void initialize(Color startingColor, Boolean visible) {
            GridBagLayout gridLayout = new GridBagLayout();
            GridBagConstraints gc = new GridBagConstraints();
            setLayout(gridLayout);
            gc.gridx = 0;
            for (int j = 0; j < 6; j += 2) {
                JPanel separator = new JPanel();
                separator.setPreferredSize(new Dimension(0, (j == 2 ? 16 : 10)));
                gc.gridy = j;
                gridLayout.setConstraints(separator, gc);
                add(separator);
            }
            gc.gridy = 0;
            for (int j = 0; j < 4; j += 2) {
                JPanel separator = new JPanel();
                separator.setPreferredSize(new Dimension(10, 0));
                gc.gridx = j;
                gridLayout.setConstraints(separator, gc);
                add(separator);
            }
            gc.gridx = 1;
            gc.gridy = 1;
            gc.anchor = GridBagConstraints.WEST;
            gc.fill = GridBagConstraints.HORIZONTAL;
            gridLayout.setConstraints(numberRange, gc);
            add(numberRange);
            colorSwatch = new ColorSwatchPanel(startingColor);
            visibilityPanel = new ToggleDetailPanel(this, "Show using color:", false,
                                                    colorSwatch, visible);
            gc.gridy = 3;
            gridLayout.setConstraints(visibilityPanel, gc);
            add(visibilityPanel);
        }
    }

    /**
     * The change range dialog content pane, allowing an
     * arbitrary range to be created by the user.
     */
    private class CheckboxTitledPanel extends PanelTitledPanel {

        public CheckboxTitledPanel(JCheckBox checkBox, JPanel panel) {
            initialize(checkBox, 5, panel);
        }
    }


    // Private Static Constants

    /**
     * List of font sizes.
     */
    private static final String[] FONT_SIZES = { "6", "7", "8", "9", "10", "11", "12", "14",
                                                 "16", "18", "20", "22", "24", "26", "28",
                                                 "36" };


    // Private Variables

    /**
     * Flag indicating whether the panel's internal state
     * is busy updating or not.
     */
    private boolean updating = false;

    /**
     * User communicator.
     */
    private UserCommunicator userCommunicator;

    /**
     * Shapefile display properties.
     */
    private ShapefileDisplayProperties properties;

    /**
     * Display palette.
     */
    private DisplayPalette displayPalette;

    /**
     * Shapefile display property information.
     */
    private ShapefileDisplayProperties.Property property;

    /**
     * Combo box providing the properties that may be
     * chosen to govern colors.
     */
    private JComboBox colorProperty;

    /**
     * Model holding value to color mappings.
     */
    private MappingTableModel colorMappings;

    /**
     * Search document, providing the search capability
     * for the table.
     */
    protected SearchDocument searchDocument;

    /**
     * Label for the search field.
     */
    private JLabel searchLabel;

    /**
     * Search text entry field, to search for property
     * values within the table when a text property
     * is being mapped.
     */
    private JTextField searchField;

    /**
     * Table showing property values and their color
     * mappings.
     */
    private SearchableTable colorMappingTable;

    /**
     * Button used to split a property value's range.
     */
    private JButton splitButton;

    /**
     * Button used to combine a property value's ranges.
     */
    private JButton combineButton;

    /**
     * Button used to change a property value's range.
     */
    private JButton rangeButton;

    /**
     * Button used to modify a property value's color.
     */
    private JButton changeColorButton;

    /**
     * Checkbox used to toggle display of labels.
     */
    private JCheckBox showingLabels;

    /**
     * Label for the font name combo box.
     */
    private JLabel fontNameLabel;

    /**
     * Combo box showing font name choices.
     */
    private JComboBox fontName;

    /**
     * Label for the font size combo box.
     */
    private JLabel fontSizeLabel;

    /**
     * Combo box showing font size choices.
     */
    private JComboBox fontSize;

    /**
     * Checkbox used to toggle bold font style.
     */
    private JCheckBox fontBold;

    /**
     * Checkbox used to toggle italic font style.
     */
    private JCheckBox fontItalic;

    /**
     * Button used to modify the label's color.
     */
    private JButton changeFontColor;

    /**
     * Label used to show a sample label.
     */
    private JLabel sampleLabel;


    // Public Constructors

    /**
     * Construct a standard instance.
     *
     * @param userCommunicator User communicator.
     * @param shapefileName    Name of the shapefile.
     * @param properties       Shapefile display properties.
     * @param displayPalette   Display palette.
     */
    public ShapefileDisplayPropertiesContent(UserCommunicator userCommunicator,
                                             String shapefileName,
                                             ShapefileDisplayProperties properties,
                                             DisplayPalette displayPalette) {
        this.userCommunicator = userCommunicator;
        this.properties = properties;
        this.displayPalette = displayPalette;

        // Ensure that there will be proper spacing between
        // the UI components.
        int j;
        GridBagLayout gridLayout = new GridBagLayout();
        GridBagConstraints gc = new GridBagConstraints();
        setLayout(gridLayout);
        gc.gridx = 0;
        JPanel separator = null;
        for (j = 0; j < 8; j += 2) {
            separator = new JPanel();
            separator.setPreferredSize(new Dimension(0, 10));
            gc.gridy = j;
            gridLayout.setConstraints(separator, gc);
            add(separator);
        }
        gc.gridy = 0;
        for (j = 0; j < 4; j += 2) {
            separator = new JPanel();
            separator.setPreferredSize(new Dimension(10, 0));
            gc.gridx = j;
            gridLayout.setConstraints(separator, gc);
            add(separator);
        }

        // Create the action listener for the widgets.
        ActionListener actionListener = new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    if (e.getActionCommand().equals("propertySelector"))
                        propertySelected();
                    else if (e.getActionCommand().equals("splitRange"))
                        splitRange();
                    else if (e.getActionCommand().equals("combineRanges"))
                        combineRanges();
                    else if (e.getActionCommand().equals("changeRange"))
                        changeRange();
                    else if (e.getActionCommand().equals("changeColor"))
                        changePropertyColor();
                    else
                        Logger.logProblem("Unknown command received: " +
                                          e.getActionCommand());
                }
            };

        // Create the shapefile choice panel.
        GridBagLayout gridLayout2 = new GridBagLayout();
        GridBagConstraints gc2 = new GridBagConstraints();
        JPanel choicePanel = new JPanel(gridLayout2);
        JLabel label = new JLabel("Shapefile:  ");
        gc2.gridx = gc2.gridy = 0;
        gridLayout2.setConstraints(label, gc2);
        choicePanel.add(label);
        JLabel shapefileChoice = new JLabel(shapefileName);
        shapefileChoice.setBorder(BorderFactory.createLoweredBevelBorder());
        gc2.gridx = 1;
        gc2.fill = GridBagConstraints.HORIZONTAL;
        gc2.weightx = 1.0;
        gridLayout2.setConstraints(shapefileChoice, gc2);
        choicePanel.add(shapefileChoice);
        gc.gridx = 1;
        gc.gridy = 1;
        gc.fill = GridBagConstraints.HORIZONTAL;
        gc.weightx = 1.0;
        gridLayout.setConstraints(choicePanel, gc);
        add(choicePanel);

        // Create the color configuration panel and its
        // subcomponents.
        JPanel colorPanel = createColorConfigurationPanel(actionListener);
        gc.gridy = 3;
        gridLayout.setConstraints(colorPanel, gc);
        add(colorPanel);

        // Create the labeling configuration panel and
        // its subcomponents.
        CheckboxTitledPanel labelPanel = createLabelConfigurationPanel();
        gc.gridy = 5;
        gridLayout.setConstraints(labelPanel, gc);
        add(labelPanel);

        // Load the current shapefile's properties into
        // the window.
        loadShapefileProperties();
    }


    // Public Methods

    /**
     * Validate the input provided by the user.
     *
     * @param  dialog Dialog in which this content pane
     *                is embedded.
     * @return True if the input is valid, otherwise
     *         false.
     */
    public boolean validateInput(ModalDialog dialog) {
        return true;
    }


    // Private Methods

    /**
     * Create the color configuration panel.
     *
     * @param  listener Action listener that will
     *                  respond to action events.
     * @return Color configuration panel just created.
     */
    private JPanel createColorConfigurationPanel(ActionListener listener) {

        // Configure the spacing between the widgets.
        GridBagLayout gridLayout = new GridBagLayout();
        GridBagConstraints gc = new GridBagConstraints();
        JPanel colorPanel = new JPanel(gridLayout);
        colorPanel.setBorder(BorderFactory.createTitledBorder("Feature Colors"));
        gc.gridx = gc.gridy = 0;
        gc.weightx = 0.0;
        gc.weighty = 0.0;
        gc.fill = GridBagConstraints.NONE;
        JPanel separator;
        int j;
        for (j = 0; j < 14; j += 2) {
            separator = new JPanel();
            separator.setPreferredSize(new Dimension(0, (j == 0 ? 3 :
                                                         (j == 2 ? 5 :
                                                          (j == 10 ? 20 : 10)))));
            gc.gridy = j;
            gc.weighty = (j == 4 ? 1.0 : 0.0);
            gridLayout.setConstraints(separator, gc);
            colorPanel.add(separator);
        }
        gc.gridy = 0;
        for (j = 0; j < 10; j += 2) {
            separator = new JPanel();
            separator.setPreferredSize(new Dimension((j == 6 ? 5 : 10), 0));
            gc.gridx = j;
            gc.fill = (j == 4 ? GridBagConstraints.HORIZONTAL : GridBagConstraints.NONE);
            gc.weightx = (j == 4 ? 1.0 : 0.0);
            gridLayout.setConstraints(separator, gc);
            colorPanel.add(separator);
        }

        // Add the labels, the search field, and the combo
        // box allowing the choosing of a property to use
        // as the determining variable. The search field
        // is given a keystroke listener so that it can
        // respond to up- and down-arrow clicks by moving
        // the selection to the previous or next item in
        // the table, respectively.
        JLabel label = new JLabel("Determining Property:    ");
        gc.gridx = gc.gridy = 1;
        gridLayout.setConstraints(label, gc);
        colorPanel.add(label);
        gc.gridx = 3;
        label = new JLabel("Colors for Values:");
        gc.gridx = 3;
        gc.anchor = GridBagConstraints.WEST;
        gridLayout.setConstraints(label, gc);
        colorPanel.add(label);
        searchLabel = new JLabel("Find:");
        gc.gridx = 5;
        gridLayout.setConstraints(searchLabel, gc);
        colorPanel.add(searchLabel);
        searchField = new JTextField(20);
        searchField.addKeyListener(new KeyAdapter() {
                public void keyPressed(KeyEvent e) {
                    if ((e.getKeyCode() == KeyEvent.VK_UP) ||
                        (e.getKeyCode() == KeyEvent.VK_DOWN)) {
                        int index = colorMappings.getSelectedIndex();
                        index += (e.getKeyCode() == KeyEvent.VK_UP ? -1 : 1);
                        if ((index < 0) || (index >= colorMappings.getSize()))
                            return;
                        colorMappings.setSelectedIndex(index);
                    }
                }
            });
        gc.gridx = 7;
        gridLayout.setConstraints(searchField, gc);
        colorPanel.add(searchField);
        colorProperty = new JComboBox();
        colorProperty.setMaximumRowCount(20);
        colorProperty.setActionCommand("propertySelector");
        colorProperty.addActionListener(listener);
        gc.gridx = 1;
        gc.gridy = 3;
        gc.anchor = GridBagConstraints.NORTH;
        gc.fill = GridBagConstraints.HORIZONTAL;
        gridLayout.setConstraints(colorProperty, gc);
        colorPanel.add(colorProperty);

        // Add the table that allows the viewing and modify-
        // ing of the value to color mappings.
        colorMappings = new MappingTableModel();
        colorMappingTable = new SearchableTable(colorMappings);
        colorMappings.setTable(colorMappingTable);
        searchDocument = new SearchDocument(null, colorMappingTable, searchField);
        searchField.setDocument(searchDocument);
        colorMappingTable.setColumnSelectionAllowed(false);
        int width = colorMappingTable.getFontMetrics(colorMappingTable.getFont()).
            stringWidth("XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX");
        colorMappingTable.getColumnModel().getColumn(0).setPreferredWidth(width);
        int totalWidth = width;
        width = colorMappingTable.getFontMetrics(colorMappingTable.getFont()).
            stringWidth("XXXXXXX");
        colorMappingTable.getColumnModel().getColumn(1).setPreferredWidth(width);
        totalWidth += width;
        width = colorMappingTable.getFontMetrics(colorMappingTable.getFont()).
            stringWidth("XXXXXXX");
        colorMappingTable.getColumnModel().getColumn(2).setPreferredWidth(width);
        colorMappingTable.getColumnModel().getColumn(2).
            setCellRenderer(new ColorTableCellRenderer(4, 4));
        totalWidth += width;
        int totalHeight = colorMappingTable.getRowHeight() * 10;
        Dimension size = new Dimension(totalWidth, totalHeight);
        colorMappingTable.setPreferredScrollableViewportSize(size);
        JScrollPane scrollPane = new JScrollPane(colorMappingTable);
        scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
        gc.gridx = 3;
        gc.weightx = 1.0;
        gc.weighty = 1.0;
        gc.gridwidth = 5;
        gc.gridheight = 9;
        gc.fill = GridBagConstraints.BOTH;
        gridLayout.setConstraints(scrollPane, gc);
        colorPanel.add(scrollPane);

        // Add the buttons allowing the modification of the
        // table's content.
        splitButton = new JButton("Split");
        splitButton.setActionCommand("splitRange");
        splitButton.addActionListener(listener);
        splitButton.setEnabled(false);
        gc.gridx = 1;
        gc.gridy = 5;
        gc.weightx = gc.weighty = 0.0;
        gc.gridwidth = gc.gridheight = 1;
        gc.anchor = GridBagConstraints.SOUTH;
        gc.fill = GridBagConstraints.HORIZONTAL;
        gridLayout.setConstraints(splitButton, gc);
        colorPanel.add(splitButton);
        combineButton = new JButton("Combine");
        combineButton.setActionCommand("combineRanges");
        combineButton.addActionListener(listener);
        combineButton.setEnabled(false);
        gc.gridy = 7;
        gridLayout.setConstraints(combineButton, gc);
        colorPanel.add(combineButton);
        rangeButton = new JButton("Range...");
        rangeButton.setActionCommand("changeRange");
        rangeButton.addActionListener(listener);
        rangeButton.setEnabled(false);
        gc.gridy = 9;
        gridLayout.setConstraints(rangeButton, gc);
        colorPanel.add(rangeButton);
        changeColorButton = new JButton("Color...");
        changeColorButton.setActionCommand("changeColor");
        changeColorButton.addActionListener(listener);
        changeColorButton.setEnabled(false);
        gc.gridy = 11;
        gridLayout.setConstraints(changeColorButton, gc);
        colorPanel.add(changeColorButton);

        // Add a listener to allow table selection changes
        // to cause state changes in the table-modifying
        // buttons, and another listener to cause double-
        // clicks to trigger color changes.
        colorMappingTable.getSelectionModel().
            addListSelectionListener(new ListSelectionListener() {
                    public void valueChanged(ListSelectionEvent e) {
                        tableSelectionChanged();
                        searchField.setText((String) colorMappings.getSelectedItem());
                        searchDocument.itemSelected();
                    }
                });
        colorMappingTable.addMouseListener(new MouseAdapter() {
                public void mouseClicked(MouseEvent e) {
                    if (e.getClickCount() == 2) {
                        Point p = e.getPoint();
                        int row = colorMappingTable.rowAtPoint(p);
                        int col = colorMappingTable.columnAtPoint(p);
                        if ((row != -1) && (col != 1)) {
                            ListSelectionModel selectionModel =
                                colorMappingTable.getSelectionModel();
                            selectionModel.addSelectionInterval(row, row);
                            changePropertyColor();
                        }
                    }
                }
            });

        // Return the finished product.
        return colorPanel;
    }

    /**
     * Create the label configuration panel.
     *
     * @return Label configuration panel just created.
     */
    private CheckboxTitledPanel createLabelConfigurationPanel() {

        // Create the action listener to respond
        // to the font name or size being chosen.
        ActionListener actionListener = new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    if (e.getSource() == fontName) {
                        properties.setFontName((String) fontName.getSelectedItem());
                        updateLabelWidgets();
                    } else if (e.getSource() == fontSize) {
                        int size = 0;
                        try {
                            size = Integer.parseInt((String) fontSize.getSelectedItem());
                        } catch (Exception e2) {
                            Logger.logBug("SHOULDN'T HAPPEN.", e2);
                            return;
                        }
                        properties.setFontSize(size);
                        updateLabelWidgets();
                    } else if (e.getSource() == changeFontColor) {
                        Color color = properties.getFontColor();
                        color = userCommunicator.
                            getUserColor(getDialogParent(ShapefileDisplayPropertiesContent.
                                                         this), color);
                        if (color != null) {
                            properties.setFontColor(color);
                            updateLabelWidgets();
                        }
                    }
                }
            };

        // Create the item listener to respond to
        // the checkbox states changing.
        ItemListener itemListener = new ItemListener() {
                public void itemStateChanged(ItemEvent e) {
                    if (e.getItemSelectable() == fontBold)
                        properties.setFontBold(e.getStateChange() == ItemEvent.SELECTED);
                    else if (e.getItemSelectable() == fontItalic)
                        properties.setFontItalic(e.getStateChange() == ItemEvent.SELECTED);
                    else
                        properties.setShowingLabels(e.getStateChange() == ItemEvent.SELECTED);
                    updateLabelWidgets();
                }
            };

        // Create the detail panel to hold all the
        // subcomponents, and configure the spacing
        // between the latter.
        GridBagLayout gridLayout = new GridBagLayout();
        GridBagConstraints gc = new GridBagConstraints();
        JPanel detailPanel = new JPanel(gridLayout);
        gc.gridx = gc.gridy = 0;
        gc.weightx = 1.0;
        gc.weighty = 0.0;
        gc.anchor = GridBagConstraints.WEST;
        gc.fill = GridBagConstraints.NONE;
        JPanel separator = new JPanel();
        separator.setPreferredSize(new Dimension(0, 10));
        gc.gridy = 1;
        gridLayout.setConstraints(separator, gc);
        detailPanel.add(separator);

        // Create the first line of the detail panel.
        JPanel nameAndSizePanel = new JPanel();
        nameAndSizePanel.setLayout(new BoxLayout(nameAndSizePanel, BoxLayout.X_AXIS));
        fontNameLabel = new JLabel("Font:");
        nameAndSizePanel.add(fontNameLabel);
        nameAndSizePanel.add(Box.createRigidArea(new Dimension(5, 0)));
        String[] names = GraphicsEnvironment.getLocalGraphicsEnvironment().
            getAvailableFontFamilyNames();
        fontName = new JComboBox(names);
        fontName.setSelectedItem(properties.getFontName());
        fontName.addActionListener(actionListener);
        nameAndSizePanel.add(fontName);
        nameAndSizePanel.add(Box.createRigidArea(new Dimension(12, 0)));
        fontSizeLabel = new JLabel("Size:");
        nameAndSizePanel.add(fontSizeLabel);
        nameAndSizePanel.add(Box.createRigidArea(new Dimension(5, 0)));
        fontSize = new JComboBox(FONT_SIZES);
        fontSize.setSelectedItem(Integer.toString(properties.getFontSize()));
        fontSize.addActionListener(actionListener);
        nameAndSizePanel.add(fontSize);
        nameAndSizePanel.add(Box.createRigidArea(new Dimension(10, 0)));
        fontBold = new JCheckBox("Bold", properties.isFontBold());
        fontBold.addItemListener(itemListener);
        nameAndSizePanel.add(fontBold);
        nameAndSizePanel.add(Box.createRigidArea(new Dimension(5, 0)));
        fontItalic = new JCheckBox("Italic", properties.isFontItalic());
        fontItalic.addItemListener(itemListener);
        nameAndSizePanel.add(fontItalic);
        nameAndSizePanel.add(Box.createRigidArea(new Dimension(7, 0)));
        changeFontColor = new JButton("Color...");
        changeFontColor.addActionListener(actionListener);
        nameAndSizePanel.add(changeFontColor);
        gc.gridy = 0;
        gridLayout.setConstraints(nameAndSizePanel, gc);
        detailPanel.add(nameAndSizePanel);
        int maxWidth = nameAndSizePanel.getPreferredSize().width;

        // Create the second line of the detail panel.
        Dimension size = new Dimension(maxWidth, 60);
        JPanel samplePanel = new JPanel(new BorderLayout());
        samplePanel.setBorder(BorderFactory.createLoweredBevelBorder());
        sampleLabel = new JLabel("");
        sampleLabel.setBackground(displayPalette.getBackgroundColor());
        sampleLabel.setOpaque(true);
        sampleLabel.setHorizontalAlignment(SwingConstants.CENTER);
        sampleLabel.setMinimumSize(size);
        sampleLabel.setMaximumSize(size);
        sampleLabel.setPreferredSize(size);
        samplePanel.add(sampleLabel, BorderLayout.CENTER);
        gc.gridy = 2;
        gridLayout.setConstraints(samplePanel, gc);
        detailPanel.add(samplePanel);
        updateLabelWidgets();

        // Create the label panel that holds the above-
        // created detail panel, and is titled with a
        // checkbox enabling or disabling the detail
        // panel's subcomponents.
        showingLabels = new JCheckBox("Show Annotation Labels",
                                      properties.isShowingLabels());
        showingLabels.addItemListener(itemListener);
        CheckboxTitledPanel labelPanel = new CheckboxTitledPanel(showingLabels, detailPanel);

        // Return the finished product.
        return labelPanel;
    }

    /**
     * Update the label widgets.
     */
    private void updateLabelWidgets() {

        // Update the sample label's look.
        boolean labelsShowing = properties.isShowingLabels();
        int style = (properties.isFontBold() ? Font.BOLD : 0) +
            (properties.isFontItalic() ? Font.ITALIC : 0);
        int size = properties.getFontSize();
        sampleLabel.setText(labelsShowing ? "Sample text" : "");
        sampleLabel.setFont(new Font(properties.getFontName(), style,
                                     properties.getFontSize()));
        sampleLabel.setForeground(properties.getFontColor());

        // Enable or disable all label widgets based
        // upon whether or not labels are being used.
        fontNameLabel.setEnabled(labelsShowing);
        fontName.setEnabled(labelsShowing);
        fontSizeLabel.setEnabled(labelsShowing);
        fontSize.setEnabled(labelsShowing);
        fontBold.setEnabled(labelsShowing);
        fontItalic.setEnabled(labelsShowing);
        changeFontColor.setEnabled(labelsShowing);
    }

    /**
     * Load shapefile properties.
     */
    private void loadShapefileProperties() {

        // Update the properties combo box and table.
        updating = true;
        colorProperty.removeAllItems();
        for (int j = 0; j < properties.getNumberOfPropertyMappings(); j++)
            colorProperty.addItem(properties.getPropertyMapping(j).getName());
        int index = properties.getCurrentPropertyMappingsIndex();
        colorProperty.setSelectedIndex(index);
        if (index != -1) {
            property = properties.getPropertyMapping(index);
            colorMappings.populate(property.getDisplayablePropertyValues(),
                                   property.getVisibilities(), property.getColors());
        }
        boolean enableSearch = ((index != -1) &&
                                (property instanceof ShapefileDisplayProperties.TextProperty));
        searchLabel.setEnabled(enableSearch);
        searchField.setEnabled(enableSearch);
        searchDocument.setModel(enableSearch ? colorMappings : null);
        tableSelectionChanged();
        updating = false;
    }

    /**
     * Respond to the user's selection of a different
     * property.
     */
    private void propertySelected() {

        // If already programmatically updating, do
        // nothing.
        if (updating)
            return;

        // Make sure that a color property was selected.
        int index = colorProperty.getSelectedIndex();
        if (index == -1)
            return;

        // Set the current property mapping to match that
        // selected, and repopulate the table.
        properties.setCurrentPropertyMapping(index);
        property = properties.getPropertyMapping(index);
        colorMappings.populate(property.getDisplayablePropertyValues(),
                               property.getVisibilities(), property.getColors());
        boolean enableSearch = ((property instanceof ShapefileDisplayProperties.TextProperty));
        searchLabel.setEnabled(enableSearch);
        searchField.setEnabled(enableSearch);
        searchDocument.setModel(enableSearch ? colorMappings : null);
        tableSelectionChanged();
    }

    /**
     * Split one or more ranges.
     */
    private void splitRange() {

        // Split any range that can be split into two (for
        // integer properties, a range can only be split
        // if it encompasses more than 1; for floating
        // ranges, any range can be split).
        int[] rows = colorMappingTable.getSelectedRows();
        ArrayList<Integer> rowList = new ArrayList<Integer>();
        if (property instanceof ShapefileDisplayProperties.IntegerProperty) {
            ShapefileDisplayProperties.IntegerProperty intProperty =
                (ShapefileDisplayProperties.IntegerProperty) property;
            int numAdded = 0;
            for (int j = 0; j < rows.length; j++) {
                int min = intProperty.getRangeBoundary(rows[j] + numAdded);
                int max = intProperty.getRangeBoundary(rows[j] + numAdded + 1);
                if (max - min > 1) {
                    Boolean visible = intProperty.isVisible(rows[j] + numAdded);
                    Color color = intProperty.getColor(rows[j] + numAdded);
                    min = ((max - min) / 2) + min;
                    intProperty.createNewRange(min, max, visible, color);
                    rowList.add(rows[j] + numAdded);
                    rowList.add(rows[j] + numAdded + 1);
                    numAdded++;
                } else
                    rowList.add(rows[j] + numAdded);
            }
        } else {
            ShapefileDisplayProperties.FloatingProperty floatProperty =
                (ShapefileDisplayProperties.FloatingProperty) property;
            for (int j = 0; j < rows.length; j++) {
                double min = floatProperty.getRangeBoundary(rows[j] + j);
                double max = floatProperty.getRangeBoundary(rows[j] + j + 1);
                Boolean visible = floatProperty.isVisible(rows[j] + j);
                Color color = floatProperty.getColor(rows[j] + j);
                min = ((max - min) / 2.0) + min;
                floatProperty.createNewRange(min, max, visible, color);
                rowList.add(rows[j] + j);
                rowList.add(rows[j] + j + 1);
            }
        }

        // Update the table to reflect the changes, keeping
        // all the split ranges selected.
        ListSelectionModel selectionModel = colorMappingTable.getSelectionModel();
        selectionModel.setValueIsAdjusting(true);
        colorMappings.populate(property.getDisplayablePropertyValues(),
                               property.getVisibilities(), property.getColors());
        selectionModel.clearSelection();
        for (int j = 0; j < rowList.size(); j++)
            selectionModel.addSelectionInterval(rowList.get(j), rowList.get(j));
        selectionModel.setValueIsAdjusting(false);
    }

    /**
     * Combine two or more ranges.
     */
    private void combineRanges() {

        // Combine all the ranges that are selected (and
        // any between selected ranges) into one range.
        int[] rows = colorMappingTable.getSelectedRows();
        Boolean visible = property.isVisible(rows[0]);
        Color color = property.getColor(rows[0]);
        if (property instanceof ShapefileDisplayProperties.IntegerProperty) {
            ShapefileDisplayProperties.IntegerProperty intProperty =
                (ShapefileDisplayProperties.IntegerProperty) property;
            int min = intProperty.getRangeBoundary(rows[0]);
            int max = intProperty.getRangeBoundary(rows[rows.length - 1] + 1);
            intProperty.createNewRange(min, max, visible, color);
        } else {
            ShapefileDisplayProperties.FloatingProperty floatProperty =
                (ShapefileDisplayProperties.FloatingProperty) property;
            double min = floatProperty.getRangeBoundary(rows[0]);
            double max = floatProperty.getRangeBoundary(rows[rows.length - 1] + 1);
            floatProperty.createNewRange(min, max, visible, color);
        }

        // Update the table to reflect the changes, keeping
        // the single range that is a combination of the
        // previously selected ranges selected.
        ListSelectionModel selectionModel = colorMappingTable.getSelectionModel();
        selectionModel.setValueIsAdjusting(true);
        colorMappings.populate(property.getDisplayablePropertyValues(),
                               property.getVisibilities(), property.getColors());
        selectionModel.clearSelection();
        selectionModel.addSelectionInterval(rows[0], rows[0]);
        selectionModel.setValueIsAdjusting(false);
    }

    /**
     * Change one or more ranges.
     */
    private void changeRange() {

        // Put up a dialog allowing the user to create a
        // new range, then add that range.  If the range
        // overruns one or more previous ranges, those
        // are automatically replaced.
        int[] rows = colorMappingTable.getSelectedRows();
        Boolean visible = property.isVisible(rows[0]);
        Color color = property.getColor(rows[0]);
        if (property instanceof ShapefileDisplayProperties.IntegerProperty) {
            ShapefileDisplayProperties.IntegerProperty intProperty =
                (ShapefileDisplayProperties.IntegerProperty) property;
            int min = intProperty.getRangeBoundary(rows[0]);
            int max = intProperty.getRangeBoundary(rows[rows.length - 1] + 1) -
                (rows[rows.length - 1] + 1 == intProperty.getSize() ? 0 : 1);
            ChangeRangeContent content =
                new ChangeRangeContent(intProperty.getMinimum(),
                                       intProperty.getMaximum(), min, max,
                                       visible, color);
            if (userCommunicator.getUserInputViaDialog(getDialogParent(this),
                                                       "Create Range",
                                                       content)) {
                intProperty.createNewRange(content.getIntegerLower(),
                                           content.getIntegerUpper() +
                                           (content.getIntegerUpper() ==
                                            intProperty.getMaximum() ? 0 : 1),
                                           content.getVisible(),
                                           content.getColor());
                for (int j = 0; j < intProperty.getSize(); j++) {
                    if (intProperty.getRangeBoundary(j) == content.getIntegerLower()) {
                        rows[0] = j;
                        break;
                    }
                }
            } else
                return;
        } else {
            ShapefileDisplayProperties.FloatingProperty floatProperty =
                (ShapefileDisplayProperties.FloatingProperty) property;
            double min = floatProperty.getRangeBoundary(rows[0]);
            double max = floatProperty.getRangeBoundary(rows[rows.length - 1] + 1);
            ChangeRangeContent content =
                new ChangeRangeContent(floatProperty.getMinimum(),
                                       floatProperty.getMaximum(), min, max, visible, color);
            if (userCommunicator.getUserInputViaDialog(getDialogParent(this),
                                                       "Create Range",
                                                       content)) {
                floatProperty.createNewRange(content.getDoubleLower(),
                                             content.getDoubleUpper(),
                                             content.getVisible(),
                                             content.getColor());
                for (int j = 0; j < floatProperty.getSize(); j++) {
                    if (floatProperty.getRangeBoundary(j) == content.getDoubleLower()) {
                        rows[0] = j;
                        break;
                    }
                }
            } else
                return;
        }

        // Update the table to reflect the changes, selec-
        // ting the newly created range.
        ListSelectionModel selectionModel = colorMappingTable.getSelectionModel();
        selectionModel.setValueIsAdjusting(true);
        colorMappings.populate(property.getDisplayablePropertyValues(),
                               property.getVisibilities(), property.getColors());
        selectionModel.clearSelection();
        selectionModel.addSelectionInterval(rows[0], rows[0]);
        selectionModel.setValueIsAdjusting(false);
    }

    /**
     * Allow the user to change the color of the selected
     * property value(s).
     */
    private void changePropertyColor() {
        int[] rows = colorMappingTable.getSelectedRows();
        Color color = property.getColor(rows[0]);
        color = userCommunicator.getUserColor(getDialogParent(this), color);
        if (color != null) {
            ListSelectionModel selectionModel = colorMappingTable.getSelectionModel();
            selectionModel.setValueIsAdjusting(true);
            for (int j = 0; j < rows.length; j++)
                property.setColor(rows[j], color);
            colorMappings.populate(property.getDisplayablePropertyValues(),
                                   property.getVisibilities(), property.getColors());
            selectionModel.clearSelection();
            for (int j = 0; j < rows.length; j++)
                selectionModel.addSelectionInterval(rows[j], rows[j]);
            selectionModel.setValueIsAdjusting(false);
        }
    }

    /**
     * Respond to a visibility checkbox being toggled in
     * the table.
     *
     * @param row     Row where toggle occurred.
     * @param visible New visiblity.
     */
    private void propertyVisibilityChanged(int row, Boolean visible) {
        property.setVisibility(row, visible);
    }

    /**
     * Respond to notification that the table's selection
     * has changed.
     */
    private void tableSelectionChanged() {
        int[] rows = colorMappingTable.getSelectedRows();
        boolean selectionExists = ((rows != null) && (rows.length > 0));
        changeColorButton.setEnabled(selectionExists);
        if (property instanceof ShapefileDisplayProperties.TextProperty) {
            splitButton.setEnabled(false);
            combineButton.setEnabled(false);
            rangeButton.setEnabled(false);
        } else if (property instanceof ShapefileDisplayProperties.IntegerProperty) {
            if (selectionExists) {
                ShapefileDisplayProperties.IntegerProperty intProperty =
                    (ShapefileDisplayProperties.IntegerProperty) property;
                boolean selectedRangeExists = false;
                for (int j = 0; j < rows.length; j++) {
                    if (intProperty.getRangeBoundary(rows[j] + 1) -
                        intProperty.getRangeBoundary(rows[j]) > 1) {
                        selectedRangeExists = true;
                        break;
                    }
                }
                splitButton.setEnabled(selectedRangeExists);
                combineButton.setEnabled(rows.length > 1);
            } else {
                splitButton.setEnabled(false);
                combineButton.setEnabled(false);
            }
            rangeButton.setEnabled(selectionExists);
        } else {
            splitButton.setEnabled(selectionExists);
            combineButton.setEnabled(selectionExists && (rows.length > 1));
            rangeButton.setEnabled(selectionExists);
        }
    }

    /**
     * Get the parent dialog.
     *
     * @return Parent dialog.
     */
    private Component getDialogParent(Component component) {
        Component parent = null;
        for (parent = component.getParent(); parent != null; parent = parent.getParent())
            if (parent instanceof ModalDialog)
                break;
        if (parent == null)
            Logger.logBug("No ModalDialog ancestor of content pane found.",
                          new Exception());
        return parent;
    }
}

