package fsl.depict.shapefile;

import fsl.depict.ProductVisualInfo;
import fsl.util.VariableStringParser;

/**
 * Shapefile product visual info, encapsulating the visual
 * information determining the basic display parameters
 * to be used for shapefile depictables for a particular
 * product.
 *
 * @author Christopher Golden
 */
public class ShapefileVisualInfo extends ProductVisualInfo {


    // Private Variables

    /**
     * List of constant strings, into which are interleaved
     * the values of the variables specified in the <code>
     * variables</code> list. This list is always one item
     * longer than the latter.
     */
    private String[] constants;

    /**
     * List of variable names; the values of these are
     * interleaved between the corresponding constant
     * strings within the <code>constants</code> list. If
     * there are no variable names, then this is <code>
     * null</code>. Otherwise, the final string is
     * assembled like this, assuming c[x] is the constant
     * at index x and v[x] is the variable at index x,
     * with n being the number of variables:
     * c[0]v[0]c[1]v[1]...c[n]v[n]c[n+1].
     */
    private String[] variables;


    // Public Constructors

    /**
     * Construct a standard instance.
     *
     * @param info String providing the starting values for
     *             this object's info.
     */
    public ShapefileVisualInfo(String info) {
        String[][] result = VariableStringParser.parseStringWithVariableSpecifiers(info);
        constants = result[0];
        variables = result[1];
    }


    // Public Methods

    /**
     * Get the list of constant strings. These constants
     * are have the values of the variables specified by
     * the list returned by the </code>getVariables()</code>
     * method interleaved between them to form the string
     * to be used as an annotation, assembled as follows,
     * assuming c[x] is the constant at index x and v[x] is
     * the variable at index x, with n being the number of
     * variables: c[0]v[0]c[1]v[1]...c[n]v[n]c[n+1].
     *
     * @return List of constant strings.
     */
    public String[] getConstants() {
        return constants;
    }

    /**
     * Get the list of variable specifiers.
     *
     * @return List of variable specifiers, or <code>null
     *         </code> if there are no such specifiers.
     */
    public String[] getVariables() {
        return variables;
    }

    /**
     * Determine whether or not this object is equivalent to
     * the specified object.
     *
     * @param  object Object to be tested against this one
     *                for equivalence.
     * @return True if the two objects are equivalent, false
     *         otherwise.
     */
    public boolean equals(Object object) {
        if (((variables == null) && (((ShapefileVisualInfo) object).variables != null)) ||
            ((variables != null) && (((ShapefileVisualInfo) object).variables == null)) ||
            ((variables != null) &&
             (variables.length != ((ShapefileVisualInfo) object).variables.length)) ||
            (constants.length != ((ShapefileVisualInfo) object).constants.length))
            return false;
        if (variables != null)
            for (int j = 0; j < variables.length; j++)
                if (variables[j].equals(((ShapefileVisualInfo) object).variables[j]) == false)
                    return false;
        for (int j = 0; j < constants.length; j++)
            if (constants[j].equals(((ShapefileVisualInfo) object).constants[j]) == false)
                return false;
        return true;
    }

    /**
     * Get the hash code of this object.
     *
     * @return Hash code of this object.
     */
    public int hashCode() {
        return (int) (((variables != null ? (long) variables.hashCode() : 0L) +
                       (long) constants.hashCode()) % (long) Integer.MAX_VALUE);
    }
}
