 # -----------------------------------------------------------------------------
# This software is in the public domain, furnished "as is", without technical 
# support, and with no warranty, express or implied, as to its usefulness for 
# any purpose.
#
# PuffVolcanoListVACTSubset.py
#
# Author: Jim Frimel
#
# MODULE:
# -------
# This class is reponsible for the creation of the puff Volcano VACT 
# Subset List file used by our FXC VACT Customers 
#
# This module can be imported or run as a stand alone program.
# To run as a stand alone program: python PuffVolcanoListVACTSubset.py
#
# ASSUMPTIONS:
# ------------
# 1. All Volcano lines except line one are of the following format. ie)
#    Volcano	   Region	    Lat	       Lon	   Elev.
#    CLEVELAND     :Aleutian Is     :52.82 :N :169.95 :W : 1730
# 2. Line one of the file is a Header Line
# 3. The loop used to generate the subset file searches for 
#    strings in the Volcano line by the Region (ie. "Aleutian Is").
#    
# INPUT: 
# ------
# FXC_HOME/fsl/data/config/puffVolcanos.txt, 
# the Original Volcano List generated by puff.
# 
# OUTPUT:
# -------
# FXC_HOME/fsl/data/config/puffVolcanosVACTSubset.txt, 
# subset of puffVolcans.txt
#
# CONFIGURABLES: 
# --------------
# To Change ... go to the init method, the constructor
#     1. Input File location and name
#     2. Output File location and name
#     3. Regions List to be used to create the VACT Subset
# 
# VERSION HISTORY:
# ----------------
# 02.07.2005  Jim Frimel	Original Version
#
#
# IMPLEMENTATION:
# ---------------
# myNotes: 
# Python Learning Developer Notes
# self.VarName => makes that attribute (variable) accessable in other methods
# self.__VarName => private, which I don't think means much in python
# ----------------------------------------------------------------------------
from os import environ 
from os.path import isfile
from string import count
import string
import sys

class PuffVolcanoListVACTSubset:
    #--------------------------------------------------------------------------
    # The initialization method, the contructor.
    #--------------------------------------------------------------------------
    def __init__(self):
        #-------------------
        # CONFIGURABLE ITEMS
        #-------------------        
        #Regions used to build the VACT Subset list
        #For a list of regions vi the puffVolcanos.txt file.
        self.volcanoRegionsVACT = ["Kurile", \
                                   "Kamchatka", \
                                   "Aleutian Is", \
                                   "Alaska Peninsula", \
                                   "Alaska-SW", \
                                   "Alaska-SE", \
                                   "Alaska-W", \
                                   "Alaska-E"]

        #The input File
	self.inFileName =  environ['FXC_HOME'] + "/fsl/data/config/" + \
	                   "puffVolcanos.txt"

        #The output File
	self.outFileName =  environ['FXC_HOME'] + "/fsl/data/config/" + \
	                    "puffVolcanosVACTSubset.txt"

        #---------------------

        #self.getSystemInfo()
        self.readPuffVolcanoListInputFile()
        self.createPuffVolcanoVACTSubsetList()
        self.writePuffVolcanoVACTSubsetListOutputFile()

    #--------------------------------------------------------------------------
    #Not needed but just getting some system information
    #--------------------------------------------------------------------------
    def getSystemInfo(self):
        try:
            systemNameInfo = environ["HOST"]
            sysNameInfo = string.split(systemNameInfo, ".")
            # bearcat, etc...
            systemName = string.upper(sysNameInfo[0])  
            #self.__systemLocation = string.upper(sysNameInfo[1]) # FSL
            self.userName = environ["USER"]            
            print "Machine: " + systemName + " " + "User: " + self.userName
        except:
            msg = "EXCEPTION: method = getSystemInfo \n" \
                  "           in PuffVolcanoListVACTSubset.py \n" \
                  "TYPE: " + str(sys.exc_type) + "\n"  \
                  "VALUE: " + str(sys.exc_value) + "\n"
            print msg

    #--------------------------------------------------------------------------
    # Reads in the Puff Volcano List file line by line
    # and creates an array of strings, a python List.
    #--------------------------------------------------------------------------
    def readPuffVolcanoListInputFile(self):
        try:
	    input = open(self.inFileName, 'r')
	    self.puffVolcanoList = input.readlines()
	    input.close()
        except:
            msg = "EXCEPTION: method = readPuffVolcanoListInputFile \n" \
                  "           in PuffVolcanoListVACTSubset.py \n" \
                  "TYPE: " + str(sys.exc_type) + "\n"  \
                  "VALUE: " + str(sys.exc_value) + "\n"
            print msg 


    #--------------------------------------------------------------------------
    #From the the Puff Volcano List create the VACT Subset List.
    #Element 0 of the VACT Subset List is the Header
    #--------------------------------------------------------------------------
    def createPuffVolcanoVACTSubsetList(self):
        try:
            #Create the VACT Subset List and assign it the Header
            self.puffVolcanoVACTSubsetList = [self.puffVolcanoList[0]]
            #Loop through all volcanoes line by line
            for volcano in self.puffVolcanoList:
                #search each line to see if it is in the region we want.
                for region in self.volcanoRegionsVACT:
                    # returns number of occurences of region in volcano line"
                    # 0 indicates the volcano is not in the region 
                    if count(volcano,region):
                        self.puffVolcanoVACTSubsetList.append(volcano)
                        break #once found exit innerloop and get another volcano.

            #For debugging ... Print out the volcano list
            #for volcano in self.puffVolcanoVACTSubsetList:
            #    print volcano,
            msg = "The number of volcanos in the VACT subset is: " + \
                   str(len(self.puffVolcanoVACTSubsetList) - 1) + "\n" 
            print msg
        
        except:
            msg = "EXCEPTION: method = createPuffVolcanoVACTSubsetList \n" \
                  "           in PuffVolcanoListVACTSubset.py \n" \
                  "TYPE: " + str(sys.exc_type) + "\n"  \
                  "VALUE: " + str(sys.exc_value) + "\n"
            print msg 

    #--------------------------------------------------------------------------
    # Writes the puffVolcanosVACTSubset.txt file
    #--------------------------------------------------------------------------
    def writePuffVolcanoVACTSubsetListOutputFile(self):
        try:
            
            if isfile(self.outFileName):
               msg = "WARNING: FILE ALREADY EXISTS!! DID NOT CREATE OUTPUT FILE. \n" + \
                     self.outFileName + "\n" \
                     "You must first delete the output file. \n"
               print msg
            else:    
	       message = open(self.outFileName, 'w')
	       message.writelines(self.puffVolcanoVACTSubsetList)
	       message.close()
               print "CREATED: " + self.outFileName + "\n"	    
	    
	except:
            msg = "EXCEPTION: method = writePuffVolcanoVACTSubsetListOutputFile \n" \
                  "           in PuffVolcanoListVACTSubset.py \n" \
                  "TYPE: " + str(sys.exc_type) + "\n"  \
                  "VALUE: " + str(sys.exc_value) + "\n"
            print msg

if __name__ == '__main__': PuffVolcanoListVACTSubset()
            
